#!/usr/bin/env python

#******************************************************************************
#
#  rpnCalendar.py
#
#  rpnChilada calendar operators
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import calendar

from convertdate import bahai, french_republican, gregorian, hebrew, indian_civil, \
                        islamic, julian, mayan, persian
from ethiopian_date import EthiopianDateConverter as ethiopian_date
from mpmath import ceil

from rpn.rpnDateTime import RPNDateTime
from rpn.rpnName import getOrdinalName
from rpn.rpnUtils import oneArgFunctionEvaluator, validateRealInt


#******************************************************************************
#
#  calendar names
#
#******************************************************************************

bahaiYears = [
    'Alif',
    'Ba\'',
    'Ab',
    'Dal',
    'Bab',
    'Vav',
    'Abad',
    'Jad',
    'Baha\'',
    'Hubb',
    'Bahhaj',
    'Javab',
    'Ahad',
    'Vahhab',
    'Vidad',
    'Badi`',
    'Bahi',
    'Abha',
    'Vahid',
]

bahaiMonths = [
    'Baha',
    'Jalal',
    'Jamal',
    '`Azamat',
    'Nur',
    'Rahmat',
    'Kalimat',
    'Kamal',
    'Asma\'',
    '`Izzat',
    'Mashiyyat',
    '`Ilm',
    'Qudrat',
    'Qawl',
    'Masa\'il',
    'Sharaf',
    'Sultan',
    'Mulk',
    'Ayyam-i-Ha',
    '`Ala\''
]

bahaiDays = [
    'Kamal',     # Monday
    'Fidal',
    '`Idal',
    'Istijlal',
    'Istiqlal',
    'Jalal',
    'Jamal'
]

hebrewMonths = [
    'Nisan',
    'Iyar',
    'Sivan',
    'Tammuz',
    'Av',
    'Elul',
    'Tishrei',
    'Marcheshvan',
    'Kislev',
    'Tevet',
    'Shevat',
    'Adar I',
    'Adar II'
]

hebrewDays = [
    'Yom Sheni',     # Monday
    'Yom Shlishi',
    'Yom Revi\'i',
    'Yom Chamishi',
    'Yom Shishi',
    'Yom Shabbat',
    'Yom Rishon'
]

indianCivilDays = [
    'Somavara',     # Monday
    'Mangalavara',
    'Budhavara',
    'Guruvara',
    'Sukravara',
    'Sanivara',
    'Ravivara',
]

indianCivilMonths = [
    'Chaitra',
    'Vaishakha',
    'Jyeshtha',
    'Ashadha',
    'Shravana',
    'Bhaadra',
    'Ashwin',
    'Kartika',
    'Agrahayana',
    'Pausha',
    'Magha',
    'Phalguna',
]

islamicMonths = [
    'Muharram',
    'Safar',
    'Rabi al-Awwal',
    'Rabi ath-Thani',
    'Jumada al-Ula',
    'Jumada ath-Thaniyah',
    'Rajab',
    'Sha`ban',
    'Ramadan',
    'Shawwal',
    'Dhu al-Qa`dah',
    'Dhu al-Hijjah'
]

islamicDays = [
    'al-Ithnayn',   # Monday
    'ath-Thulatha',
    'al-Arbi`a',
    'al-Khamis',
    'al-Jumu`ah',
    'as-Sabt',
    'al-Ahad'
]

persianMonths = [
    'Farvardin',
    'Ordibehesht',
    'Khordad',
    'Tir',
    'Mordad',
    'Shahrivar',
    'Mehr',
    'Aban',
    'Azar',
    'Dey',
    'Bahman',
    'Esfand'
]

persianDays = [
    'Yekshanbeh',   # Monday
    'Doshanbeh',
    'Seshhanbeh',
    'Chaharshanbeh',
    'Panjshanbeh',
    'Jomeh',
    'Shanbeh'
]

ethiopianMonths = [
    'Meskerem',
    'Thikimt',
    'Hidar',
    'Tahsas',
    'Thir',
    'Yekatit',
    'Megabit',
    'Miyazya',
    'Ginbot',
    'Senie',
    'Hamlie',
    'Nehasie',
    'Phagumien'
]

ethiopianDays = [
    'Lideta',
    'Aba Guba',
    'Be\'Eta',
    'Yohannes',
    'Abo',
    'Iyesus',
    'Selassie',
    'Aba Kiros',
    'Tomas',
    'Meskel',
    'Hana Mariam',
    'Michael',
    'Egziher Ab',
    'Abune Aregawi',
    'K\'irk\'os',
    'Kidane Mihret',
    'Est\'ifanos',
    'Tekle Alfa',
    'Gabriel',
    'Hintsite',
    'Mariam',
    'Ura\'el',
    'Giorgis',
    'Tekle Haimanot',
    'Merk\'orios',
    'Yosef',
    'Medhani Alem',
    'Amanu\'el',
    'Bale Egziabeir',
    'Mark\'os'
]

frenchRepublicanDays = [
    'Primidi',
    'Duodi',
    'Tridi',
    'Quartidi',
    'Quintidi',
    'Sextidi',
    'Septidi',
    'Octidi',
    'Nonidi',
    'Decadi'
]

frenchRepublicanMonths = [
    'Vendemiaire',
    'Brumaire',
    'Frimaire',
    'Nivose',
    'Pluviose',
    'Ventose',
    'Germinal',
    'Floreal',
    'Prairial',
    'Messidor',
    'Thermidor',
    'Fructidor',
]


#******************************************************************************
#
#  getOrdinalDate
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getOrdinalDate( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'a date-time type required for this operator' )

    return str( n.year ) + '-' + str( n.timetuple( ).tm_yday )


#******************************************************************************
#
#  generateMonthCalendar
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def generateMonthCalendar( datetime ):
    if not isinstance( datetime, RPNDateTime ):
        raise ValueError( 'a date-time type required for this operator' )

    cal = calendar.TextCalendar( firstweekday = 6 )
    print( )
    cal.prmonth( datetime.year, datetime.month )
    print( )

    return [ datetime.year, datetime.month ]


#******************************************************************************
#
#  generateYearCalendar
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def generateYearCalendar( n ):
    cal = calendar.TextCalendar( firstweekday = 6 )

    if isinstance( n, RPNDateTime ):
        cal.pryear( n.year )
    else:
        cal.pryear( validateRealInt( n ) )

    return ''


#******************************************************************************
#
#  getJulianDay
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getJulianDay( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'a date-time type required for this operator' )

    return gregorian.to_jd( n.year, n.month, n.day )


#******************************************************************************
#
#  getLilianDay
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getLilianDay( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'a date-time type required for this operator' )

    return ceil( n.subtract( RPNDateTime( 1582, 10, 15 ) ).value )


#******************************************************************************
#
#  getISODate
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getISODate( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'a date-time type required for this operator' )

    result = n.isocalendar( )

    return list( result )


#******************************************************************************
#
#  getISODateName
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getISODateName( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'a date-time type required for this operator' )

    result = n.isocalendar( )

    return str( result[ 0 ] ) + '-W' + str( result[ 1 ] ) + '-' + str( result[ 2 ] )


#******************************************************************************
#
#  getHebrewCalendarDate
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getHebrewCalendarDate( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    return list( hebrew.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertHebrewDate
#
#******************************************************************************

def convertHebrewDate( year, month, day ):
    return RPNDateTime( *hebrew.to_gregorian( validateRealInt( year ), validateRealInt( month ),
                                              validateRealInt( day ) ), dateOnly = True )


#******************************************************************************
#
#  getHebrewCalendarDateName
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getHebrewCalendarDateName( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    date = hebrew.from_gregorian( n.year, n.month, n.day )

    return hebrewDays[ n.weekday( ) ] + ', ' + hebrewMonths[ date[ 1 ] - 1 ] + \
           ' ' + str( date[ 2 ] ) + ', ' + str( date[ 0 ] )


#******************************************************************************
#
#  getIndianCivilCalendarDate
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getIndianCivilCalendarDate( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    return list( indian_civil.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertIndianCivilDate
#
#******************************************************************************

def convertIndianCivilDate( year, month, day ):
    return RPNDateTime( *indian_civil.to_gregorian( validateRealInt( year ), validateRealInt( month ),
                                                    validateRealInt( day ) ),
                        dateOnly = True )


#******************************************************************************
#
#  getIndianCivilCalendarDateName
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getIndianCivilCalendarDateName( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    date = indian_civil.from_gregorian( n.year, n.month, n.day )

    return indianCivilDays[ n.weekday( ) ] + ', ' + indianCivilMonths[ date[ 1 ] - 1 ] + \
           ' ' + str( date[ 2 ] ) + ', ' + str( date[ 0 ] )


#******************************************************************************
#
#  getMayanCalendarDate
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getMayanCalendarDate( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    return list( mayan.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertMayanDate
#
#******************************************************************************

def convertMayanDate( baktun, katun, tun, uinal, kin ):
    return RPNDateTime( *mayan.to_gregorian( validateRealInt( baktun ), validateRealInt( katun ),
                                             validateRealInt( tun ), validateRealInt( uinal ),
                                             validateRealInt( kin ) ), dateOnly = True )


#******************************************************************************
#
#  getIslamicCalendarDate
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getIslamicCalendarDate( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    return list( islamic.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertIslamicDate
#
#******************************************************************************

def convertIslamicDate( year, month, day ):
    return RPNDateTime( *islamic.to_gregorian( validateRealInt( year ), validateRealInt( month ),
                                               validateRealInt( day ) ),
                        dateOnly = True )


#******************************************************************************
#
#  getIslamicCalendarDateName
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getIslamicCalendarDateName( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    date = islamic.from_gregorian( n.year, n.month, n.day )

    return islamicDays[ n.weekday( ) ] + ', ' + islamicMonths[ date[ 1 ] - 1 ] + \
           ' ' + str( date[ 2 ] ) + ', ' + str( date[ 0 ] )


#******************************************************************************
#
#  getJulianCalendarDate
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getJulianCalendarDate( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    return list( julian.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertJulianDate
#
#******************************************************************************

def convertJulianDate( year, month, day ):
    return RPNDateTime( *julian.to_gregorian( validateRealInt( year ), validateRealInt( month ),
                                              validateRealInt( day ) ),
                        dateOnly = True )


#******************************************************************************
#
#  getPersianCalendarDate
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getPersianCalendarDate( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    return list( persian.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertPersianDate
#
#******************************************************************************

def convertPersianDate( year, month, day ):
    return RPNDateTime( *persian.to_gregorian( validateRealInt( year ), validateRealInt( month ),
                                               validateRealInt( day ) ),
                        dateOnly = True )


#******************************************************************************
#
#  getPersianCalendarDateName
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getPersianCalendarDateName( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    date = persian.from_gregorian( n.year, n.month, n.day )

    return persianDays[ n.weekday( ) ] + ', ' + persianMonths[ date[ 1 ] - 1 ] + \
           ' ' + str( date[ 2 ] ) + ', ' + str( date[ 0 ] )


#******************************************************************************
#
#  getBahaiCalendarDate
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getBahaiCalendarDate( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    return list( bahai.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertBahaiDate
#
#******************************************************************************

def convertBahaiDate( year, month, day ):
    return RPNDateTime( *bahai.to_gregorian( int( year ), int( month ), int( day ) ), dateOnly = True )


#******************************************************************************
#
#  getBahaiCalendarDateName
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getBahaiCalendarDateName( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    date = bahai.from_gregorian( n.year, n.month, n.day )

    result = bahaiDays[ n.weekday( ) ] + ', ' + bahaiMonths[ date[ 1 ] - 1 ] + \
           ' ' + str( date[ 2 ] ) + ', '

    if date[ 0 ] >= 1:
        result += 'Year ' + bahaiYears[ date[ 0 ] % 19 - 1 ] + ' of the ' + \
                  getOrdinalName( ( date[ 0 ] // 19 ) + 1 ) + ' Vahid of the ' + \
                  getOrdinalName( ( date[ 0 ] // 361 ) + 1 ) + ' Kull-i-Shay\'' + \
                  ' (Year ' + str( date[ 0 ] ) + ')'
    else:
        result += str( date[ 0 ] )

    return result


#******************************************************************************
#
#  getEthiopianCalendarDate
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getEthiopianCalendarDate( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    return list( ethiopian_date.to_ethiopian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertEthiopianDate
#
#******************************************************************************

def convertEthiopianDate( year, month, day ):
    ethDate = ethiopian_date.to_gregorian( validateRealInt( year ), validateRealInt( month ), validateRealInt( day ) )
    return RPNDateTime( ethDate.year, ethDate.month, ethDate.day )


#******************************************************************************
#
#  getEthiopianCalendarDateName
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getEthiopianCalendarDateName( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    date = list( ethiopian_date.to_ethiopian( n.year, n.month, n.day ) )

    return ethiopianDays[ date[ 2 ] - 1 ] + ' ' + ethiopianMonths[ date[ 1 ] - 1 ] + \
                          ' ' + str( date[ 0 ] )


#******************************************************************************
#
#  getFrenchRepublicanCalendarDate
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getFrenchRepublicanCalendarDate( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    return list( french_republican.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertFrenchRepublicanDate
#
#******************************************************************************

def convertFrenchRepublicanDate( year, month, day ):
    return RPNDateTime( *french_republican.to_gregorian( int( year ), int( month ), int( day ) ), dateOnly = True )


#******************************************************************************
#
#  getFrenchRepublicanCalendarDateName
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getFrenchRepublicanCalendarDateName( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'time type required for this operator' )

    date = french_republican.from_gregorian( n.year, n.month, n.day )

    return frenchRepublicanDays[ n.weekday( ) ] + ', ' + frenchRepublicanMonths[ date[ 1 ] - 1 ] + \
           ' ' + str( date[ 2 ] ) + ', ' + str( int( date[ 0 ] ) )





#!/usr/bin/env python

#******************************************************************************
#
#  rpnDice.py
#
#  rpnChilada dice simulation operators
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import itertools
import re

from collections import Counter
from random import randrange

from mpmath import arange

from rpn.rpnDebug import debugPrint
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnUtils import oneArgFunctionEvaluator, twoArgFunctionEvaluator


# TODO:  write a function to enumerate the number of outcomes for a dice expression
#        without actually permuting them.  i.e., rpn <dice_expr> permute_dice occurrences
#
#        Is that possible?  It's definitely possible without including x and h.  With them
#        it becomes a problem that's harder than I can solve, probably.
#
# c:\>rpn [ 1 6 dup ] 3 poly_power 3 enumerate -s1
# [
#     [ 3, 1 ],
#     [ 4, 3 ],
#     [ 5, 6 ],
#     [ 6, 10 ],
#     [ 7, 15 ],
#     [ 8, 21 ],
#     [ 9, 25 ],
#     [ 10, 27 ],
#     [ 11, 27 ],
#     [ 12, 25 ],
#     [ 13, 21 ],
#     [ 14, 15 ],
#     [ 15, 10 ],
#     [ 16, 6 ],
#     [ 17, 3 ],
#     [ 18, 1 ]
# ]
#
# c:\>rpn 3d6 permute_dice occurrences -s1
# [
#     [ 3, 1 ],
#     [ 4, 3 ],
#     [ 5, 6 ],
#     [ 6, 10 ],
#     [ 7, 15 ],
#     [ 8, 21 ],
#     [ 9, 25 ],
#     [ 10, 27 ],
#     [ 11, 27 ],
#     [ 12, 25 ],
#     [ 13, 21 ],
#     [ 14, 15 ],
#     [ 15, 10 ],
#     [ 16, 6 ],
#     [ 17, 3 ],
#     [ 18, 1 ]
# ]
#
# Make a new data type (dict) and then write an operator for summing multiple dicts.  Piece o' cake.
#
#


#******************************************************************************
#
#  rollDice
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def rollDice( expression ):
    values, modifier = evaluateDiceExpression( parseDiceExpression( expression ) )
    return sum( values ) + modifier


#******************************************************************************
#
#  rollMultipleDice
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def rollMultipleDice( expression, times ):
    dice = parseDiceExpression( expression )

    for _ in arange( 0, times ):
        values, modifier = evaluateDiceExpression( dice )
        yield sum( values ) + modifier

def rollMultipleDiceGenerator( n, k ):
    return RPNGenerator( rollMultipleDice( n, k ) )


#******************************************************************************
#
#  enumerateDice
#
#******************************************************************************

def enumerateDice( expression ):
    return evaluateDiceExpression( parseDiceExpression( expression ), False )[ 0 ]

@oneArgFunctionEvaluator( )
def enumerateDiceGenerator( n ):
    return RPNGenerator( enumerateDice( n ) )


#******************************************************************************
#
#  enumerateMultipleDice
#
#******************************************************************************

def enumerateMultipleDice( expression, count ):
    dice = parseDiceExpression( expression )

    for _ in arange( 0, count ):
        yield evaluateDiceExpression( dice )[ 0 ]

@twoArgFunctionEvaluator( )
def enumerateMultipleDiceGenerator( n, k ):
    return RPNGenerator( enumerateMultipleDice( n, k ) )


#******************************************************************************
#
#  permuteDice
#
#  format: [c]dv[x[p]][h[q]][-+]y
#
#  c - dice count, defaults to 1
#  v - dice value, i.e., number of sides, minumum 2
#  p - drop lowest die value(s), defaults to 1
#  q - drop highest value(s), defaults to 1
#  y - add or subtract y from the total (modifier)
#
#******************************************************************************

def permuteDice( expression ):
    dice = parseDiceExpression( expression )

    diceList = [ ]

    dropAny = False
    ranges = [ ]
    dropRanges = [ ]
    drops = [ ]
    start = 0
    modifierTotal = 0
    groupCount = len( dice )

    for diceCount, diceValue, dropLowest, dropHighest, modifier in dice:
        ranges.append( ( start, start + diceCount ) )
        start += diceCount

        if dropLowest > 0 or dropHighest > 0:
            drops.append( True )
            dropRanges.append( ( dropLowest, diceCount - dropHighest ) )
            dropAny = True
        else:
            drops.append( False )
            dropRanges.append( ( 0, 0 ) )

        modifierTotal += modifier

        for i in range( diceCount ):
            diceList.append( range( 1, diceValue + 1 ) )

    permutations = itertools.product( *diceList )

    for values in permutations:
        if dropAny:
            total = 0

            for i in range( 0, groupCount ):
                group = values[ ranges[ i ][ 0 ] : ranges[ i ][ 1 ] ]

                if drops[ i ]:
                    group = list( group )
                    group.sort( )
                    total += sum( group[ dropRanges[ i ][ 0 ] : dropRanges[ i ][ 1 ] ] )
                else:
                    total += sum( group )

                total += modifierTotal
                yield total
        else:
            yield sum( values ) + modifierTotal

@oneArgFunctionEvaluator( )
def permuteDiceGenerator( n ):
    return RPNGenerator( permuteDice( n ) )


#******************************************************************************
#
#  parseDiceExpression
#
#  format: [c]dv[,[c]dv...][x[p]][h[q]][-+]y
#
#  c - dice count, defaults to 1
#  v - dice value, i.e., number of sides, minumum 2
#  p - drop lowest die value(s), defaults to 1
#  q - drop highest value(s), defaults to 1
#  y - add or subtract y from the total (modifier)
#
#  return values:
#
#  a list of 5-tuples:  dice count, dice value, drop lowest count,
#                       drop highest count, modifier
#
#******************************************************************************

def parseDiceExpression( arg ):
    counter = Counter( arg )

    if counter[ 'a' ] > 1:
        raise ValueError( 'dice expressions can only contain a single \'x\'' )

    if counter[ 'h' ] > 1:
        raise ValueError( 'dice expressions can only contain a single \'h\'' )

    if counter[ '+' ] > 1:
        raise ValueError( 'dice expressions can only contain a single \'+\'' )

    if counter[ '-' ] > 1:
        raise ValueError( 'dice expressions can only contain a single \'-\'' )

    if counter[ '+' ] + counter[ '-' ] > 1:
        raise ValueError( 'dice expressions can only have a single modifier (\'+\' or \'-\')' )

    expressions = re.split( ',', arg )

    result = [ ]

    for expr in expressions:
        pieces = re.split( '([dhx+-])', expr )

        debugPrint( 'pieces', pieces )

        diceValue = 0
        diceCount = 0
        dropLowestCount = 0
        dropHighestCount = 0
        modifier = 0

        defaultState = 0
        diceValueState = 1
        dropLowestCountState = 2
        dropHighestCountState = 3
        plusModifierState = 4
        minusModifierState = 5

        state = defaultState

        for piece in pieces:
            if state == defaultState:
                if piece == 'd':
                    state = diceValueState
                elif piece == 'x':
                    state = dropLowestCountState
                elif piece == 'h':
                    state = dropHighestCountState
                elif piece == '-':
                    state = minusModifierState
                elif piece == '+':
                    state = plusModifierState
                else:
                    if piece == '':
                        diceCount = 1
                    else:
                        diceCount = int( piece )
            elif state == diceValueState:
                diceValue = int( piece )

                if diceValue < 2:
                    raise ValueError( 'dice value must be greater than 1' )

                state = defaultState
            elif state == dropLowestCountState:
                if piece == '':
                    dropLowestCount = 1
                else:
                    dropLowestCount = int( piece )

                if dropLowestCount < 1:
                    raise ValueError( 'drop lowest count must be positive' )

                state = defaultState
            elif state == dropHighestCountState:
                if piece == '':
                    dropHighestCount = 1
                else:
                    dropHighestCount = int( piece )

                if dropHighestCount < 1:
                    raise ValueError( 'drop highest count must be positive' )

            elif state == plusModifierState:
                if piece == '':
                    modifier = 1
                else:
                    modifier = int( piece )

                state = defaultState
            elif state == minusModifierState:
                if piece == '':
                    modifier = -1
                else:
                    modifier = -int( piece )

                state = defaultState

        if diceCount == 0:
            diceCount = 1

        if dropLowestCount != 0 and diceValue == 0:
            raise ValueError( 'dice expression requires \'d\' if \'x\' is used' )

        if dropHighestCount != 0 and diceValue == 0:
            raise ValueError( 'dice expression requires \'d\' if \'h\' is used' )

        if ( dropLowestCount + dropHighestCount ) >= diceCount:
            raise ValueError( 'this dice expression is dropping as many or more dice than are being rolled' )

        debugPrint( 'expression', expr )
        debugPrint( 'diceCount', diceCount )
        debugPrint( 'diceValue', diceValue )
        debugPrint( 'dropLowestCount', dropLowestCount )
        debugPrint( 'dropHighestCount', dropHighestCount )
        debugPrint( 'modifier', modifier )

        result.append( ( diceCount, diceValue, dropLowestCount, dropHighestCount, modifier ) )

    return result


#******************************************************************************
#
#  evaluateDiceExpression
#
#  a list of 5-tuples:  dice count, dice value, drop lowest count,
#                       drop highest count, modifier
#
#******************************************************************************

def evaluateDiceExpression( args, sumIfPossible=True ):
    result = [ ]
    modifierSum = 0    # currently only a single modifier is allowed

    if sumIfPossible:
        result = [ 0 ]

    for diceCount, diceValue, dropLowestCount, dropHighestCount, modifier in args:
        modifierSum += modifier

        if dropLowestCount == 0 and dropHighestCount == 0:
            if sumIfPossible:
                for _ in range( 0, diceCount ):
                    result[ 0 ] += ( randrange( diceValue ) + 1 )
            else:
                for _ in range( 0, diceCount ):
                    result.append( randrange( diceValue ) + 1 )
        else:
            dice = [ ]

            for _ in range( 0, diceCount ):
                dice.append( randrange( diceValue ) + 1 )

            dice.sort( )
            debugPrint( 'dice', dice )

            if dropHighestCount > 0:
                debugPrint( 'drop', dice[ dropLowestCount : -dropHighestCount ] )
                result.extend( dice[ dropLowestCount : -dropHighestCount ] )
            else:
                debugPrint( 'drop', dice[ dropLowestCount : ] )
                result.extend( dice[ dropLowestCount : ] )

    return result, modifierSum


#******************************************************************************
#
#  rollSimpleDice
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def rollSimpleDice( n, k ):
    values, modifier = evaluateDiceExpression( [ ( int( n ), int( k ), 0, 0, 0 ) ] )
    return sum( values ) + modifier




#!/usr/bin/env python

#******************************************************************************
#
#  rpnNanoseconds
#
#  nanosecond-resolution timer for Python <= 3.6 on Linux
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import ctypes


#******************************************************************************
#
#  time_ns
#
#  https://stackoverflow.com/questions/55774054/precise-time-in-nano-seconds-for-python-3-6-and-earlier
#
#  This is a Linux-specific implementation, but I've never needed this for
#  Windows since I can always control what version of Python I am running.
#
#******************************************************************************

CLOCK_REALTIME = 0

class TimeSpec( ctypes.Structure ):
    _fields_ = [
        ( 'tv_sec', ctypes.c_int64 ), # seconds, https://stackoverflow.com/q/471248/1672565
        ( 'tv_nsec', ctypes.c_int64 ), # nanoseconds
    ]

clock_gettime = ctypes.cdll.LoadLibrary( 'libc.so.6' ).clock_gettime
clock_gettime.argtypes = [ ctypes.c_int64, ctypes.POINTER( TimeSpec ) ]
clock_gettime.restype = ctypes.c_int64

def time_ns( ):
    timespec = TimeSpec( )
    ret = clock_gettime( CLOCK_REALTIME, ctypes.pointer( timespec ) )

    if bool( ret ):
        raise OSError( )

    return timespec.tv_sec * 1_000_000_000 + timespec.tv_nsec




#!/usr/bin/env python

#******************************************************************************
#
#  rpnGenerator.py
#
#  rpnChilada generator object declarations
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import itertools

from mpmath import arange, ceil, fadd, fdiv, fmul, fsub, mpmathify, power


#******************************************************************************
#
#  class RPNGenerator
#
#******************************************************************************

class RPNGenerator( ):
    '''This class implements generators for rpn values, which allows rpn to
    avoid having to expand ranges to complete lists when evaluating them.'''
    def __init__( self, generator = None, count = -1 ):
        self.generator = generator
        self.count = count

    def getGenerator( self ):
        return self.generator

    def getCount( self ):
        return self.count

    def __iter__( self ):
        return self.generator

    def __next__( self ):
        return self.generator.__next__( )

    def __getitem__( self, index ):
        if isinstance( index, slice ):
            return next( itertools.islice( self.generator, index.start, index.end, 1 ) )
        else:
            return next( itertools.islice( self.generator, index, index + 1 ) )

    def clone( self ):
        self.generator, newGenerator = itertools.tee( self.generator )
        return RPNGenerator( newGenerator, self.count )

    @staticmethod
    def create( value ):
        if isinstance( value, RPNGenerator ):
            return value
        elif isinstance( value, list ):
            count = len( value )
        else:
            count = 1

        return RPNGenerator( itemGenerator( value ), count )

    @staticmethod
    def createRange( start, end, step = 1 ):
        try:
            # attempt to calculate the number of items to be generated
            result = RPNGenerator( rangeGenerator( start, end, step ),
                                   ceil( fdiv( fadd( fsub( end, start ), 1 ), step ) ) )
        except TypeError:
            # if that fails, then just do without the count
            result = RPNGenerator( rangeGenerator( start, end, step ) )

        return result

    @staticmethod
    def createSizedRange( start, step = 1, count = 1 ):
        end = fadd( start, fmul( fsub( count, 1 ), step ) )
        return RPNGenerator( rangeGenerator( start, end, step ), count )

    @staticmethod
    def createGeometric( value, step, count ):
        return RPNGenerator( geometricRangeGenerator( value, step, count ), count )

    @staticmethod
    def createExponential( value, step, count ):
        return RPNGenerator( exponentialRangeGenerator( value, step, count ), count )

    @staticmethod
    def createChained( generator, func ):
        return RPNGenerator( chainedGenerator( generator, func ) )

    @staticmethod
    def createFilter( generator, func, invert=False ):
        return RPNGenerator( filterGenerator( generator, func, invert ) )

    @staticmethod
    def createPermutations( value ):
        return RPNGenerator( permutationGenerator( value ) )

    @staticmethod
    def createStringProduct( value ):
        return RPNGenerator( stringProductGenerator( value ) )

    @staticmethod
    def createProduct( value ):
        return RPNGenerator( productGenerator( value ) )

    @staticmethod
    def createGenerator( func, value ):
        if isinstance( value, list ):
            return RPNGenerator( func( *value ) )
        else:
            return RPNGenerator( func( value ) )


#******************************************************************************
#
#  itemGenerator
#
#  A generator for a single item or list, used in list operators
#
#******************************************************************************

def itemGenerator( value ):
    if isinstance( value, list ):
        yield from value
    else:
        yield from [ value ]


#******************************************************************************
#
#  rangeGenerator
#
#******************************************************************************

def rangeGenerator( start, end, step ):
    if start > end and step > 0:
        step = -step

    current = start

    if step > 0:
        while current <= end:
            yield current
            current = fadd( current, step )
    else:
        while current >= end:
            yield current
            current = fadd( current, step )


#******************************************************************************
#
#  geometricRangeGenerator
#
#******************************************************************************

def geometricRangeGenerator( value, step, count ):
    current = value

    for _ in arange( 0, count ):
        yield current
        current = fmul( current, step )


#******************************************************************************
#
#  exponentialRangeGenerator
#
#******************************************************************************

def exponentialRangeGenerator( value, step, count ):
    current = value

    for _ in arange( 0, count ):
        yield current
        current = power( current, step )


#******************************************************************************
#
#  evaluateNestedArgs
#
#******************************************************************************

def evaluateNestedArgs( func, i ):
    if isinstance( i, ( list, RPNGenerator ) ):
        return [ evaluateNestedArgs( func, j ) for j in i ]
    else:
        return func( i )


#******************************************************************************
#
#  chainedGenerator
#
#******************************************************************************

def chainedGenerator( generator, func ):
    for i in generator:
        yield evaluateNestedArgs( func, i )


#******************************************************************************
#
#  filterGenerator
#
#******************************************************************************

def filterGenerator( generator, func, invert=False ):
    for i in generator:
        if func( i ) != invert:
            yield i


#******************************************************************************
#
#  permutationGenerator
#
#******************************************************************************

def permutationGenerator( value ):
    for permutation in itertools.permutations( value ):
        yield mpmathify( ''.join( permutation ) )


#******************************************************************************
#
#  stringProductGenerator
#
#******************************************************************************

def stringProductGenerator( value ):
    for product in itertools.product( *value ):
        yield mpmathify( ''.join( product ) )


#******************************************************************************
#
#  productGenerator
#
#******************************************************************************

def productGenerator( value ):
    for item in itertools.product( *value ):
        yield [ mpmathify( i ) for i in item ]






#******************************************************************************
#
#  rpnPrimeUtils.py
#
#  rpnChilada prime number utilies
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import os
import sys

from bisect import bisect_left
from pathlib import Path

from mpmath import arange, fadd, fmod, fmul, fprod, fsub

import gmpy2

from rpn.rpnDebug import debugPrint
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnPrimes import primes
from rpn.rpnPersistence import cachedFunction, openPrimeCache
from rpn.rpnUtils import getUserDataPath, oneArgFunctionEvaluator, twoArgFunctionEvaluator, \
                         validateRealInt

import rpn.rpnGlobals as g


#******************************************************************************
#
#  isPrime
#
#******************************************************************************

def isPrime( n ):
    if g.zhangConjecturesAllowed and n < 1543267864443420616877677640751301:
        return isPrimeMillerRabin( int( n ) )
    elif n < 3317044064679887385961981:
        return isPrimeMillerRabin( int( n ) )
    else:
        debugPrint( 'primality testing of ' + str( int( n ) ) )
        return 1 if gmpy2.is_bpsw_prp( int( n ) ) else 0

@oneArgFunctionEvaluator( )
def isComposite( n ):
    return 0 if n == 1 or isPrime( n ) else 1

@oneArgFunctionEvaluator( )
def isPrimeOperator( n ):
    return 1 if isPrime( n ) else 0


#******************************************************************************
#
#  isPrimeNumberSimple
#
#  Use this for numbers smaller than 10,000,000,000.
#
#******************************************************************************

def isPrimeNumberSimple( n ):
    # if prime is already in the list, just pick it
    if n <= max( primes ):
        i = bisect_left( primes, n )
        return i != len( primes ) and primes[ i ] == n

    # Divide by each known prime
    limit = int( n ** .5 )

    for p in primes:
        if p > limit:
            return True

        if n % p == 0:
            return False

    return True


#******************************************************************************
#
#  getNextPrimeCandidateForAny
#
#  p is the next prime candidate.  Starting with ( p - 10 ) mod 210, we only
#  need to check 43 values out of the next 210, which eliminates all multiples
#  of 2, 3, 5, and 7 and saves us a lot of unnecessary checking.
#
#  We only need to check for a prime if ( p - 10 ) mod 210 is one of the
#  following:
#
#  1, 3, 7, 9, 13, 19, 21, 27, 31, 33, 37, 43, 49, 51, 57, 61, 63, 69, 73,
#  79, 87, 91, 93, 97, 99, 103, 111, 117, 121, 127, 129, 133, 139, 141, 147,
#  153, 157, 159, 163, 169, 171, 177, 181, 183, 187, 189, 199, 201
#
#  This list can be generated by:
#
#  rpn 10 4 primorial 10 + range lambda x 5 prime is_rough filter 10 -
#
#******************************************************************************

nextPrimeModuloTable = [
    # 0
    1,
    # 1
    2, 1,
    # 3
    4, 3, 2, 1,
    # 7
    2, 1,
    # 9
    4, 3, 2, 1,
    # 13
    6, 5, 4, 3, 2, 1,
    # 19
    2, 1,
    # 21
    6, 5, 4, 3, 2, 1,
    # 27
    4, 3, 2, 1,
    # 31
    2, 1,
    # 33
    4, 3, 2, 1,
    # 37
    6, 5, 4, 3, 2, 1,
    # 43
    6, 5, 4, 3, 2, 1,
    # 49
    2, 1,
    # 51
    6, 5, 4, 3, 2, 1,
    # 57
    4, 3, 2, 1,
    # 61
    2, 1,
    # 63
    6, 5, 4, 3, 2, 1,
    # 69
    4, 3, 2, 1,
    # 73
    6, 5, 4, 3, 2, 1,
    # 79
    8, 7, 6, 5, 4, 3, 2, 1,
    # 87
    4, 3, 2, 1,
    # 91
    2, 1,
    # 93
    4, 3, 2, 1,
    # 97
    2, 1,
    # 99
    4, 3, 2, 1,
    # 103
    8, 7, 6, 5, 4, 3, 2, 1,
    # 111
    6, 5, 4, 3, 2, 1,
    # 117
    4, 3, 2, 1,
    # 121
    6, 5, 4, 3, 2, 1,
    # 127
    2, 1,
    # 129
    4, 3, 2, 1,
    # 133
    6, 5, 4, 3, 2, 1,
    # 139
    2, 1,
    # 141
    6, 5, 4, 3, 2, 1,
    # 147
    6, 5, 4, 3, 2, 1,
    # 153
    4, 3, 2, 1,
    # 157
    2, 1,
    # 159
    4, 3, 2, 1,
    # 163
    6, 5, 4, 3, 2, 1,
    # 169
    2, 1,
    # 171
    6, 5, 4, 3, 2, 1,
    # 177
    4, 3, 2, 1,
    # 181
    2, 1,
    # 183
    4, 3, 2, 1,
    # 187
    2, 1,
    # 189
    10, 9, 8, 7, 6, 5, 4, 3, 2, 1,
    # 199
    2, 1,
    # 201
    10, 9, 8, 7, 6, 5, 4, 3, 2
]

def getNextPrimeCandidateForAny( p ):
    f = int( ( p - 10 ) % 210 )
    return p + nextPrimeModuloTable[ f ]


#******************************************************************************
#
#  getPreviousPrimeCandidateForAny
#
#  p is the next prime candidate.  Starting with ( p - 10 ) mod 210, we only
#  need to check 43 values out of the next 210, which eliminates all multiples
#  of 2, 3, 5, and 7 and saves us a lot of unnecessary checking.
#
#  We only need to check for a prime if ( p - 10 ) mod 210 is one of the
#  following:
#
#  0, 2, 4, 8, 10, 14, 20, 22, 28, 32, 34, 38, 44, 50, 52, 58, 62, 64, 70, 74
#  80, 88, 92, 94, 98, 100, 104, 112, 118, 122, 128, 130, 134, 140, 142, 148
#  154, 158, 160, 164, 170, 172, 178, 182, 184, 188, 190, 200, 202,
#
#  This list can be generated by:
#
#  rpn [ 0 ] 10 4 primorial 10 + range lambda x 5 prime is_rough filter 10 - 1 + append
#
#******************************************************************************

previousPrimeModuloTable = [
    # 0
    9, 10,
    # 2
    1, 2,
    # 4
    1, 2, 3, 4,
    # 8
    1, 2,
    # 10
    1, 2, 3, 4,
    # 14
    1, 2, 3, 4, 5, 6,
    # 20
    1, 2,
    # 22
    1, 2, 3, 4, 5, 6,
    # 28
    1, 2, 3, 4,
    # 32
    1, 2,
    # 34
    1, 2, 3, 4,
    # 38
    1, 2, 3, 4, 5, 6,
    # 44
    1, 2, 3, 4, 5, 6,
    # 50
    1, 2,
    # 52
    1, 2, 3, 4, 5, 6,
    # 58
    1, 2, 3, 4,
    # 62
    1, 2,
    # 64
    1, 2, 3, 4, 5, 6,
    # 70
    1, 2, 3, 4,
    # 74
    1, 2, 3, 4, 5, 6,
    # 80
    1, 2, 3, 4, 5, 6, 7, 8,
    # 88
    1, 2, 3, 4,
    # 92
    1, 2,
    # 94
    1, 2, 3, 4,
    # 98
    1, 2,
    # 100
    1, 2, 3, 4,
    # 104
    1, 2, 3, 4, 5, 6, 7, 8,
    # 112
    1, 2, 3, 4, 5, 6,
    # 118
    1, 2, 3, 4,
    # 122
    1, 2, 3, 4, 5, 6,
    # 128
    1, 2,
    # 130
    1, 2, 3, 4,
    # 134
    1, 2, 3, 4, 5, 6,
    # 140
    1, 2,
    # 142
    1, 2, 3, 4, 5, 6,
    # 148
    1, 2, 3, 4, 5, 6,
    # 154
    1, 2, 3, 4,
    # 158
    1, 2,
    # 160
    1, 2, 3, 4,
    # 164
    1, 2, 3, 4, 5, 6,
    # 170
    1, 2,
    # 172
    1, 2, 3, 4, 5, 6,
    # 178
    1, 2, 3, 4,
    # 182
    1, 2,
    # 184
    1, 2, 3, 4,
    # 188
    1, 2,
    # 190
    1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
    # 200
    1, 2,
    # 202
    1, 2, 3, 4, 5, 6, 7, 8,
]


#******************************************************************************
#
#  getPreviousPrimeCandidateForAny
#
#******************************************************************************

def getPreviousPrimeCandidateForAny( p ):
    f = int( ( p - 10 ) % 210 )
    return p - previousPrimeModuloTable[ f ]


#******************************************************************************
#
#  getNextPrime
#
#******************************************************************************

def getNextPrime( p, func=getNextPrimeCandidateForAny ):
    if p < 2:
        return 2

    if p == 2:
        return 3

    if p < 5:
        return 5

    if p < 7:
        return 7

    p = func( p )

    debugPrint( 'testing', p )

    while not isPrime( p ):
        p = func( p )

    return p

@oneArgFunctionEvaluator( )
@cachedFunction( 'next_prime' )
def getNextPrimeOperator( n ):
    return getNextPrime( n, func=getNextPrimeCandidateForAny )


#******************************************************************************
#
#  getPreviousPrime
#
#******************************************************************************

def getPreviousPrime( p, func=getPreviousPrimeCandidateForAny ):
    if p < 12:
        if p < 3:
            raise ValueError( 'There is no previous prime to 2.' )

        if p == 3:
            return 2

        if p <= 5:
            return 3

        if p <= 7:
            return 5

        if p <= 11:
            return 7

    p = func( p )

    while not isPrime( p ):
        p = func( p )

    return p

@oneArgFunctionEvaluator( )
@cachedFunction( 'previous_prime' )
def getPreviousPrimeOperator( n ):
    return getPreviousPrime( n, func=getPreviousPrimeCandidateForAny )


#******************************************************************************
#
#  getNextPrimes
#
#******************************************************************************

def getNextPrimes( p, k, func=getNextPrimeCandidateForAny ):
    result = [ ]

    for _ in arange( 0, k ):
        p = getNextPrime( p, func )
        result.append( p )

    return result

@twoArgFunctionEvaluator( )
def getNextPrimesOperator( n, k ):
    return getNextPrimes( n, k, func=getNextPrimeCandidateForAny )


#******************************************************************************
#
#  getPreviousPrimes
#
#******************************************************************************

def getPreviousPrimes( p, k, func = getPreviousPrimeCandidateForAny ):
    if p < 1000000000 and getNthPrime( p ) < k:
        raise ValueError( 'There is no previous prime to 2.' )

    result = [ ]

    for _ in arange( 0, k ):
        p = getPreviousPrime( p, func )
        result.append( p )

    return result

@twoArgFunctionEvaluator( )
def getPreviousPrimesOperator( n, k ):
    return getPreviousPrimes( n, k, func=getPreviousPrimeCandidateForAny )


#******************************************************************************
#
#  getNthPrime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthPrime( arg ):
    n = int( arg )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 2

    if n == 2:
        return 3

    if n == 3:
        return 5

    if g.primeDataAvailable and n >= 1000000000:
        openPrimeCache( 'huge_primes' )

        maxIndex = g.cursors[ 'huge_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'huge_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    elif g.primeDataAvailable and n >= 1000000:
        openPrimeCache( 'large_primes' )

        currentIndex, p = g.cursors[ 'large_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    elif g.primeDataAvailable and n >= 100:
        openPrimeCache( 'small_primes' )

        currentIndex, p = g.cursors[ 'small_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 4
        p = 7

    while n > currentIndex:
        p = getNextPrime( p )
        currentIndex += 1

    return p


#******************************************************************************
#
#  findPrime
#
#******************************************************************************

def findPrime( arg ):
    target = int( arg )

    if target < 3:
        return 1, 2
    elif target == 3:
        return 2, 3
    elif target <= 5:
        return 3, 5
    elif target <= 7:
        return 4, 7
    elif target < 541:          # 100th prime
        currentIndex = 4
        p = 7
    elif g.primeDataAvailable and target <= 15485863:     # 1,000,000th prime
        openPrimeCache( 'small_primes' )

        currentIndex, p = g.cursors[ 'small_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( target, ) ).fetchone( )
    elif g.primeDataAvailable and target <= 22801763489:  # 1,000,000,000th prime
        openPrimeCache( 'large_primes' )

        currentIndex, p = g.cursors[ 'large_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( target, ) ).fetchone( )
    elif g.primeDataAvailable:
        openPrimeCache( 'huge_primes' )

        currentIndex, p = g.cursors[ 'huge_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( target, ) ).fetchone( )
    else:
        currentIndex = 4
        p = 7

    while True:
        oldP = p
        p = getNextPrime( p )
        currentIndex += 1

        if p > target:
            return currentIndex - 1, oldP

@oneArgFunctionEvaluator( )
def findPrimeOperator( n ):
    return findPrime( n )[ 0 ]


#******************************************************************************
#
#  findQuadrupletPrimes
#
#******************************************************************************

def findQuadrupletPrimes( arg ):
    n = int( validateRealInt( arg ) )

    if n < 5:
        return 1, [ 5, 7, 11, 13 ]

    if n < 11:
        return 2, [ 11, 13, 17, 19 ]

    if g.primeDataAvailable:
        openPrimeCache( 'quad_primes' )

        currentIndex, p = g.cursors[ 'quad_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( n, ) ).fetchone( )
    else:
        p = 11

    while True:
        # All quadruplet primes mod 30 == 11
        p += 30

        if isPrime( p ) and isPrime( p + 2 ) and isPrime( p + 6 ) and isPrime( p + 8 ):
            currentIndex += 1

            if p > n:
                return currentIndex, [ p, p + 2, p + 6, p + 8 ]

@oneArgFunctionEvaluator( )
def findQuadrupletPrimeOperator( n ):
    return findQuadrupletPrimes( n )[ 0 ]

@oneArgFunctionEvaluator( )
def getNextQuadrupletPrime( n ):
    return findQuadrupletPrimes( n )[ 1 ][ 0 ]

@oneArgFunctionEvaluator( )
def getNextQuadrupletPrimes( n ):
    return findQuadrupletPrimes( n )[ 1 ]


#******************************************************************************
#
#  getNthQuadrupletPrime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthQuadrupletPrime( arg ):
    n = int( validateRealInt( arg ) )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 5

    if n == 2:
        return 11

    if g.primeDataAvailable and n >= 10:
        openPrimeCache( 'quad_primes' )

        maxIndex = g.cursors[ 'quad_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        startingPlace, p = g.cursors[ 'quad_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        startingPlace = 2
        p = 11

    # after 5, the first of a prime quadruplet must be a number of the form 30n + 11
    while n > startingPlace:
        p += 30

        if isPrime( p ) and isPrime( p + 2 ) and isPrime( p + 6 ) and isPrime( p + 8 ):
            n -= 1

    return p


#******************************************************************************
#
#  getNthIsolatedPrime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthIsolatedPrime( arg ):
    n = int( validateRealInt( arg ) )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 2

    if n == 2:
        return 23

    if g.primeDataAvailable:
        openPrimeCache( 'isolated_primes' )

        currentIndex, p = g.cursors[ 'isolated_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 2
        p = 23

    while n > currentIndex:
        p = getNextPrime( p )

        if not isPrime( p - 2 ) and not isPrime( p + 2 ):
            currentIndex += 1

    return p


#******************************************************************************
#
#  getNextTwinPrimeCandidate
#
#  Looking at ( p - 10 ) mod 30, the only twin prime candidates are 1, 7,
#  and 19.
#
#******************************************************************************

def getNextTwinPrimeCandidate( p ):
    f = ( p - 10 ) % 30

    if f == 1:
        p += 6
    elif f == 7:
        p += 12
    else:  # f == 19
        p += 12

    return p


#******************************************************************************
#
#  getNthTwinPrime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthTwinPrime( arg ):
    n = int( validateRealInt( arg ) )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 3

    if n == 2:
        return 5

    if n == 3:
        return 11

    if g.primeDataAvailable and n >= 100:
        openPrimeCache( 'twin_primes' )

        maxIndex = g.cursors[ 'twin_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'twin_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 3
        p = 11

    while n > currentIndex:
        p = getNextPrime( p, getNextTwinPrimeCandidate )

        if isPrime( p + 2 ):
            currentIndex += 1

    return p


#******************************************************************************
#
#  getNthTwinPrimeList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthTwinPrimeList( arg ):
    p = getNthTwinPrime( arg )
    return [ p, fadd( p, 2 ) ]


#******************************************************************************
#
#  getNthBalancedPrime
#
#  returns the first balanced prime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthBalancedPrime( arg ):
    n = int( validateRealInt( arg ) )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 5
    elif g.primeDataAvailable:
        openPrimeCache( 'balanced_primes' )

        maxIndex = g.cursors[ 'balanced_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'balanced_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )

        if n == currentIndex:
            return p

        # fill in the variables as if we'd been searching...
        secondPrevPrime = getPreviousPrime( p )
        prevPrime = p
        p = getNextPrime( prevPrime )
    else:
        currentIndex = 1
        p = 7
        prevPrime = 5
        secondPrevPrime = 3

    while n > currentIndex:
        secondPrevPrime = prevPrime
        prevPrime = p
        p = getNextPrime( p )

        if ( prevPrime - secondPrevPrime ) == ( p - prevPrime ):
            currentIndex += 1

    return prevPrime


#******************************************************************************
#
#  getNthBalancedPrimeList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthBalancedPrimeList( arg ):
    q = getNthBalancedPrime( arg )
    p = getPreviousPrime( q )
    r = getNextPrime( q )

    return [ p, q, r ]


#******************************************************************************
#
#  getNthDoubleBalancedPrimeElement
#
#******************************************************************************

def getNthDoubleBalancedPrimeElement( arg, first = False ):
    n = int( validateRealInt( arg ) )

    numberOfPrimes = 5
    center = numberOfPrimes // 2

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 18731

    if g.primeDataAvailable:
        openPrimeCache( 'double_balanced_primes' )

        maxIndex = g.cursors[ 'double_balanced_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )
        currentIndex, p = g.cursors[ 'double_balanced_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )

        if currentIndex == n:
            return p
    else:
        if n > 50:
            sys.stderr.write( 'The prime number cache data is not available.  This could take some time...\n' )

        # no cache... we'll do this the hard way
        currentIndex = 1
        p = 18731

    primes = [ 0 ] * numberOfPrimes

    primes[ center ] = p

    for i in range( 1, center + 1 ):
        primes[ center - i ] = getPreviousPrime( primes[ center - i + 1 ] )
        primes[ center + i ] = getNextPrime( primes[ center + i - 1 ] )

    while n > currentIndex:
        p = getNextPrime( primes[ -1 ] )
        primes.append( p )
        del primes[ 0 ]

        balanced = True

        for i in range( 1, center + 1 ):
            if ( primes[ center - i + 1 ] - primes[ center - i ] ) != \
               ( primes[ center + i ] - primes[ center + i - 1 ] ):
                balanced = False
                break

        if balanced:
            currentIndex += 1

    return primes[ 0 ] if first else primes[ center ]


#******************************************************************************
#
#  getNthDoubleBalancedPrimeList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthDoubleBalancedPrimeList( arg ):
    p = getNthDoubleBalancedPrimeElement( arg, first = False )

    result = [ 0, 0, p, 0, 0 ]
    middle = 2

    for i in range( 0, 2 ):
        result[ middle + i + 1 ] = getNextPrime( result[ middle + i ] )
        result[ middle - ( i + 1 ) ] = getPreviousPrime( result[ middle - i ] )

    return result

@oneArgFunctionEvaluator( )
def getNthDoubleBalancedPrime( arg ):
    return getNthDoubleBalancedPrimeElement( arg )


#******************************************************************************
#
#  getNthTripleBalancedPrimeElement
#
#******************************************************************************

def getNthTripleBalancedPrimeElement( arg, first = False ):
    n = int( validateRealInt( arg ) )

    numberOfPrimes = 7
    center = numberOfPrimes // 2

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 683783

    if g.primeDataAvailable:
        openPrimeCache( 'triple_balanced_primes' )

        maxIndex = g.cursors[ 'triple_balanced_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'triple_balanced_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )

        if n == currentIndex:
            return p
    else:
        if n > 10:
            sys.stderr.write( 'The prime number cache data is not available.  This could take some time...\n' )

        # no cache... we'll do this the hard way
        currentIndex = 1
        p = 683783

    primes = [ 0 ] * numberOfPrimes

    primes[ center ] = p

    for i in range( 1, center + 1 ):
        primes[ center - i ] = getPreviousPrime( primes[ center - i + 1 ] )
        primes[ center + i ] = getNextPrime( primes[ center + i - 1 ] )

    while n > currentIndex:
        p = getNextPrime( primes[ -1 ] )
        primes.append( p )
        del primes[ 0 ]

        balanced = True

        for i in range( 1, center + 1 ):
            if ( primes[ center - i + 1 ] - primes[ center - i ] ) != \
               ( primes[ center + i ] - primes[ center + i - 1 ] ):
                balanced = False
                break

        if balanced:
            currentIndex += 1

    return primes[ 0 ] if first else primes[ center ]


#******************************************************************************
#
#  getNthTripleBalancedPrimeList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthTripleBalancedPrimeList( arg ):
    p = getNthTripleBalancedPrimeElement( arg, first = True )

    result = [ 0, 0, 0, p, 0, 0, 0 ]
    middle = 3

    for i in range( 0, 3 ):
        result[ middle + i + 1 ] = getNextPrime( result[ middle + i ] )
        result[ middle - ( i + 1 ) ] = getPreviousPrime( result[ middle - i ] )

    return result


@oneArgFunctionEvaluator( )
def getNthTripleBalancedPrime( arg ):
    return getNthTripleBalancedPrimeElement( arg )


#******************************************************************************
#
#  getNthQuadrupleBalancedPrimeElement
#
#******************************************************************************

def getNthQuadrupleBalancedPrimeElement( arg, first = False ):
    n = int( validateRealInt( arg ) )

    numberOfPrimes = 9
    center = numberOfPrimes // 2

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 98303927

    if g.primeDataAvailable:
        openPrimeCache( 'quadruple_balanced_primes' )

        maxIndex = g.cursors[ 'quadruple_balanced_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'quadruple_balanced_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )

        if n == currentIndex:
            return p
    else:
        if n > 10:
            sys.stderr.write( 'The prime number cache data is not available.  This could take some time...\n' )

        # no cache... we'll do this the hard way
        currentIndex = 1
        p = 98303927

    primes = [ 0 ] * numberOfPrimes

    primes[ center ] = p

    for i in range( 1, center + 1 ):
        primes[ center - i ] = getPreviousPrime( primes[ center - i + 1 ] )
        primes[ center + i ] = getNextPrime( primes[ center + i - 1 ] )

    while n > currentIndex:
        p = getNextPrime( primes[ -1 ] )
        print( p )
        primes.append( p )
        del primes[ 0 ]

        balanced = True

        for i in range( 1, center + 1 ):
            if ( primes[ center - i + 1 ] - primes[ center - i ] ) != \
               ( primes[ center + i ] - primes[ center + i - 1 ] ):
                balanced = False
                break

        if balanced:
            currentIndex += 1

    return primes[ 0 ] if first else primes[ center ]


#******************************************************************************
#
#  getNthQuadrupleBalancedPrimeList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthQuadrupleBalancedPrimeList( arg ):
    p = getNthQuadrupleBalancedPrimeElement( arg, first = True )

    result = [ 0, 0, 0, 0, p, 0, 0, 0, 0 ]
    middle = 4

    for i in range( 0, 4 ):
        result[ middle + i + 1 ] = getNextPrime( result[ middle + i ] )
        result[ middle - ( i + 1 ) ] = getPreviousPrime( result[ middle - i ] )

    return result


@oneArgFunctionEvaluator( )
def getNthQuadrupleBalancedPrime( arg ):
    return getNthQuadrupleBalancedPrimeElement( arg )


#******************************************************************************
#
#  getNthSophiePrime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthSophiePrime( arg ):
    n = int( validateRealInt( arg ) )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 2

    if n == 2:
        return 3

    if n == 3:
        return 5

    if g.primeDataAvailable and n >= 100:
        openPrimeCache( 'sophie_primes' )

        maxIndex = g.cursors[ 'sophie_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'sophie_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 4
        p = 11

    while n > currentIndex:
        p = getNextPrime( p )

        if isPrime( 2 * p + 1 ):
            currentIndex += 1

    return p

@oneArgFunctionEvaluator( )
def getSafePrime( n ):
    return fadd( fmul( getNthSophiePrime( n ), 2 ), 1 )


#******************************************************************************
#
#  getNthCousinPrime
#
#  http://oeis.org/A023200
#
#  Validate: rpn 1 52 range cousin 23200 oeis  -
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthCousinPrime( arg ):
    n = int( validateRealInt( arg ) )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 3

    if g.primeDataAvailable and n >= 100:
        openPrimeCache( 'cousin_primes' )

        maxIndex = g.cursors[ 'cousin_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'cousin_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 2
        p = 7

    while n > currentIndex:
        p = getNextPrime( p )

        if isPrime( p + 4 ):
            currentIndex += 1

    return p


#******************************************************************************
#
#  getNthCousinPrimeList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthCousinPrimeList( arg ):
    p = getNthCousinPrime( arg )
    return [ p, fadd( p, 4 ) ]


#******************************************************************************
#
#  getNextSexyPrimeCandidate
#
#  For a number ( p - 10 ) mod 30, the only sexy candidates are 1, 3, 7, 13,
#  21 and 27.
#
#******************************************************************************

def getNextSexyPrimeCandidate( p ):
    f = ( p - 10 ) % 30

    if f == 1:
        p += 2
    elif f == 3:
        p += 4
    elif f == 7:
        p += 6
    elif f == 13:
        p += 8
    elif f == 21:
        p += 6
    else:   # f == 27
        p += 4

    return p


#******************************************************************************
#
#  getNthSexyPrime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthSexyPrime( arg ):
    n = int( validateRealInt( arg ) )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 5

    if g.primeDataAvailable and n >= 100:
        openPrimeCache( 'sexy_primes' )

        maxIndex = g.cursors[ 'sexy_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        startingPlace, p = g.cursors[ 'sexy_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        startingPlace = 2
        p = 7

    while n > startingPlace:
        p = getNextPrime( p, getNextSexyPrimeCandidate )

        if isPrime( p + 6 ):
            n -= 1

    return p


#******************************************************************************
#
#  getNthSexyPrimeList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthSexyPrimeList( arg ):
    p = getNthSexyPrime( arg )
    return [ p, fadd( p, 6 ) ]


#******************************************************************************
#
#  getNthSexyTriplet
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthSexyTriplet( arg ):
    n = int( validateRealInt( arg ) )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 7

    if g.primeDataAvailable and n >= 100:
        openPrimeCache( 'sexy_triplets' )

        maxIndex = g.cursors[ 'sexy_triplets' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        startingPlace, p = g.cursors[ 'sexy_triplets' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        startingPlace = 1
        p = 7

    f = p % 10

    while n > startingPlace:
        if f == 1:
            p += 6
            f = 7
        else:
            p += 4
            f = 1

        if isPrime( p ) and isPrime( p + 6 ) and isPrime( p + 12 ) and not isPrime( p + 18 ):
            n -= 1

    return p


#******************************************************************************
#
#  getNthSexyTripletList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthSexyTripletList( arg ):
    p = getNthSexyTriplet( arg )
    return [ p, fadd( p, 6 ), fadd( p, 12 ) ]


#******************************************************************************
#
#  getNthSexyQuadruplet
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthSexyQuadruplet( arg ):
    n = int( validateRealInt( arg ) )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 5

    if n < 100:
        startingPlace = 2
        p = 11
    elif g.primeDataAvailable and n >= 100:
        openPrimeCache( 'sexy_quadruplets' )

        maxIndex = g.cursors[ 'sexy_quadruplets' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        startingPlace, p = g.cursors[ 'sexy_quadruplets' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )

    while n > startingPlace:
        p += 10

        if isPrime( p ) and isPrime( p + 6 ) and isPrime( p + 12 ) and isPrime( p + 18 ):
            n -= 1

    return p


#******************************************************************************
#
#  getNthSexyQuadrupletList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthSexyQuadrupletList( arg ):
    p = getNthSexyQuadruplet( arg )
    return [ p, fadd( p, 6 ), fadd( p, 12 ), fadd( p, 18 ) ]


#******************************************************************************
#
#  getNextOctyPrimeCandidate
#
#  For a number ( p - 10 ) mod 30, the only sexy candidates are 1, 3, 9, 11,
#  13, 19, 21, 23 and 29.
#
#******************************************************************************

def getNextOctyPrimeCandidate( p ):
    f = ( p - 10 ) % 30

    if f == 1:
        p += 2
    elif f == 3:
        p += 6
    elif f == 9:
        p += 2
    elif f == 11:
        p += 2
    elif f == 13:
        p += 6
    elif f == 19:
        p += 2
    elif f == 21:
        p += 2
    elif f == 23:
        p += 6
    else:  # 29
        p += 2

    return p


#******************************************************************************
#
#  getNthOctyPrime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthOctyPrime( arg ):
    n = int( validateRealInt( arg ) )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 3

    if n == 2:
        return 5

    if g.primeDataAvailable:
        openPrimeCache( 'octy_primes' )

        maxIndex = g.cursors[ 'octy_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        startingPlace, p = g.cursors[ 'octy_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        startingPlace = 3
        p = 11

    while n > startingPlace:
        p = getNextPrime( p, getNextOctyPrimeCandidate )

        if isPrime( p + 8 ):
            n -= 1

    return p


#******************************************************************************
#
#  getNthSexyPrimeList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthOctyPrimeList( arg ):
    p = getNthOctyPrime( arg )
    return [ p, fadd( p, 8 ) ]


#******************************************************************************
#
#  getNthTripletPrimeList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthTripletPrimeList( arg ):
    n = int( validateRealInt( arg ) )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return [ 5, 7, 11 ]

    if n == 2:
        return [ 7, 11, 13 ]

    if n == 3:
        return [ 11, 13, 17 ]

    if g.primeDataAvailable and n >= 100:
        openPrimeCache( 'triplet_primes' )

        maxIndex = g.cursors[ 'triplet_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'triplet_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 3
        p = 11

    f = p % 10

    pPlus2 = False

    while n > currentIndex:
        if f == 1:
            p += 2
            f = 3
        elif f == 3:
            p += 4
            f = 7
        else:
            p += 4
            f = 1

        if isPrime( p ) and isPrime( p + 6 ):
            if isPrime( p + 2 ):
                pPlus2 = True
                currentIndex += 1
            elif isPrime( p + 4 ):
                pPlus2 = False
                currentIndex += 1

    return [ p, p + 2 if pPlus2 else p + 4, p + 6 ]


#******************************************************************************
#
#  getNthTripletPrime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthTripletPrime( arg ):
    return getNthTripletPrimeList( arg )[ 0 ]


#******************************************************************************
#
#  getNthQuadrupletPrimeList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthQuadrupletPrimeList( arg ):
    p = getNthQuadrupletPrime( arg )
    return [ p, fadd( p, 2 ), fadd( p, 6 ), fadd( p, 8 ) ]


#******************************************************************************
#
#  getNextQuintupletPrimeCandidate
#
#  For ( p - 10 ) mod 30, the only quintuplet prime candidates are:
#  1, 7, 21, or 27
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNextQuintupletPrimeCandidate( p ):
    f = p % 30

    if f == 7:
        p += 4
    else:
        p += 26

    return p


#******************************************************************************
#
#  getNthQuintupletPrime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthQuintupletPrime( arg ):
    n = int( validateRealInt( arg ) )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 5

    if n == 2:
        return 7

    if g.primeDataAvailable and n >= 10:
        openPrimeCache( 'quint_primes' )

        maxIndex = g.cursors[ 'quint_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'quint_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 3
        p = 11

    # after 5, the first of a prime quintruplet must be a number of the form 30n + 11
    while n > currentIndex:
        p = getNextPrime( p, getNextQuintupletPrimeCandidate )

        f = p % 30

        if ( ( f == 11 ) and isPrime( p + 2 ) and isPrime( p + 6 ) and \
                             isPrime( p + 8 ) and isPrime( p + 12 ) ) or \
           ( ( f == 7 ) and isPrime( p + 4 ) and isPrime( p + 6 ) and \
                            isPrime( p + 10 ) and isPrime( p + 12 ) ):
            currentIndex += 1

    return p


#******************************************************************************
#
#  getNthQuintupletPrimeList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthQuintupletPrimeList( arg ):
    if arg == 1:
        return [ 5, 7, 11, 13, 17 ]

    p = getNthQuintupletPrime( arg )

    f = p % 10

    if f == 1:
        return [ p, fadd( p, 2 ), fadd( p, 6 ), fadd( p, 8 ), fadd( p, 12 ) ]
    elif f == 7:
        return [ p, fadd( p, 4 ), fadd( p, 6 ), fadd( p, 10 ), fadd( p, 12 ) ]
    else:
        raise ValueError( 'internal error:  getNthQuintupletPrimeList is broken' )


#******************************************************************************
#
#  findQuintupletPrimes
#
#******************************************************************************

def findQuintupletPrimes( arg ):
    n = int( validateRealInt( arg ) )

    if n < 5:
        return 1, [ 5, 7, 11, 13, 17 ]
    elif n < 7:
        return 2, [ 7, 11, 13, 17, 19 ]

    if g.primeDataAvailable:
        openPrimeCache( 'quint_primes' )

        currentIndex, p = g.cursors[ 'quint_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( n, ) ).fetchone( )
    else:
        currentIndex = 3
        p = 11

    while True:
        p += 30

        f = p % 10

        if ( ( f == 1 ) and isPrime( p + 2 ) and isPrime( p + 6 ) and isPrime( p + 8 ) and isPrime( p + 12 ) ) or \
           ( ( f == 7 ) and isPrime( p + 4 ) and isPrime( p + 6 ) and isPrime( p + 10 ) and isPrime( p + 12 ) ):
            currentIndex += 1

            if p > n:
                if f == 1:
                    return currentIndex, [ p, fadd( p, 2 ), fadd( p, 6 ), fadd( p, 8 ), fadd( p, 12 ) ]
                elif f == 7:
                    return currentIndex, [ p, fadd( p, 4 ), fadd( p, 6 ), fadd( p, 10 ), fadd( p, 12 ) ]

@oneArgFunctionEvaluator( )
def findQuintupletPrimeOperator( n ):
    return findQuintupletPrimes( n )[ 0 ]

@oneArgFunctionEvaluator( )
def getNextQuintupletPrime( n ):
    return findQuintupletPrimes( n )[ 1 ][ 0 ]

@oneArgFunctionEvaluator( )
def getNextQuintupletPrimes( n ):
    return findQuintupletPrimes( n )[ 1 ]


#******************************************************************************
#
#  findSextupletPrimes
#
#******************************************************************************

def findSextupletPrimes( arg ):
    n = int( validateRealInt( arg ) )

    if n < 7:
        return 1, [ 7, 11, 13, 17, 19, 23 ]
    elif n < 97:
        return 2, [ 97, 101, 103, 107, 109, 113 ]

    if g.primeDataAvailable:
        openPrimeCache( 'sext_primes' )

        currentIndex, p = g.cursors[ 'sext_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( n, ) ).fetchone( )
    else:
        currentIndex = 3
        p = 16057

    while True:
        p += 210

        if isPrime( p + 4 ) and isPrime( p + 6 ) and isPrime( p + 10 ) and \
           isPrime( p + 12 ) and isPrime( p + 16 ):
            currentIndex += 1

            if p > n:
                return currentIndex, [ p, fadd( p, 4 ), fadd( p, 6 ), fadd( p, 10 ), fadd( p, 12 ), fadd( p, 16 ) ]

@oneArgFunctionEvaluator( )
def findSextupletPrimeOperator( n ):
    return findSextupletPrimes( n )[ 0 ]

@oneArgFunctionEvaluator( )
def getNextSextupletPrime( n ):
    return findSextupletPrimes( n )[ 1 ][ 0 ]

@oneArgFunctionEvaluator( )
def getNextSextupletPrimes( n ):
    return findSextupletPrimes( n )[ 1 ]


#******************************************************************************
#
#  findTripletPrimes
#
#******************************************************************************

def findTripletPrimes( arg ):
    n = int( validateRealInt( arg ) )

    if n < 5:
        return 1, [ 5, 7, 11 ]

    if g.primeDataAvailable:
        openPrimeCache( 'triplet_primes' )

        currentIndex, p = g.cursors[ 'triplet_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( n, ) ).fetchone( )
    else:
        currentIndex = 2
        p = 7

    while True:
        p = getNextPrime( p )

        if isPrime( p + 6 ):
            if isPrime( p + 2 ):
                currentIndex += 1

                if p > n:
                    return currentIndex, [ p, fadd( p, 2 ), fadd( p, 6 ) ]
            elif isPrime( p + 4 ):
                currentIndex += 1

                if p > n:
                    return currentIndex, [ p, fadd( p, 4 ), fadd( p, 6 ) ]

@oneArgFunctionEvaluator( )
def findTripletPrimeOperator( n ):
    return findTripletPrimes( n )[ 0 ]

@oneArgFunctionEvaluator( )
def getNextTripletPrime( n ):
    return findTripletPrimes( n )[ 1 ][ 0 ]

@oneArgFunctionEvaluator( )
def getNextTripletPrimes( n ):
    return findTripletPrimes( n )[ 1 ]



#******************************************************************************
#
#  findTwinPrimes
#
#******************************************************************************

def findTwinPrimes( arg ):
    n = int( validateRealInt( arg ) )

    if n < 5:
        return 1, [ 3, 5 ]

    if g.primeDataAvailable:
        openPrimeCache( 'twin_primes' )

        currentIndex, p = g.cursors[ 'twin_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( n, ) ).fetchone( )
    else:
        currentIndex = 2
        p = 5

    while True:
        p = getNextPrime( p )

        if isPrime( p + 2 ):
            currentIndex += 1

            if p > n:
                return currentIndex, [ p, fadd( p, 2 ) ]

@oneArgFunctionEvaluator( )
def findTwinPrimeOperator( n ):
    return findTwinPrimes( n )[ 0 ]

@oneArgFunctionEvaluator( )
def getNextTwinPrime( n ):
    return findTwinPrimes( n )[ 1 ][ 0 ]

@oneArgFunctionEvaluator( )
def getNextTwinPrimes( n ):
    return findTwinPrimes( n )[ 1 ]


#******************************************************************************
#
#  getNthSextupletPrime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthSextupletPrime( arg ):
    n = int( validateRealInt( arg ) )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 7

    if n == 2:
        return 97

    if g.primeDataAvailable:
        openPrimeCache( 'sext_primes' )

        maxIndex = g.cursors[ 'sext_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        startingPlace, p = g.cursors[ 'sext_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        startingPlace = 2
        p = 97

    # all sets of prime sextuplets must start with 210x+97
    while n > startingPlace:
        p += 210

        if isPrime( p ) and isPrime( p + 4 ) and isPrime( p + 6 ) and \
           isPrime( p + 10 ) and isPrime( p + 12 ) + isPrime( p + 16 ):
            n -= 1

    return p

@oneArgFunctionEvaluator( )
def findSextupletPrimeOperator( n ):
    return findSextupletPrimes( n )[ 0 ]

@oneArgFunctionEvaluator( )
def getNextSextupletPrime( n ):
    return findSextupletPrimes( n )[ 1 ][ 0 ]

@oneArgFunctionEvaluator( )
def getNextSextupletPrimes( n ):
    return findSextupletPrimes( n )[ 1 ]


#******************************************************************************
#
#  getNthSextupletPrimeList
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthSextupletPrimeList( arg ):
    p = getNthSextupletPrime( arg )
    return [ p, fadd( p, 4 ), fadd( p, 6 ), fadd( p, 10 ), fadd( p, 12 ), fadd( p, 16 ) ]


#******************************************************************************
#
#  getNthPrimeRange
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getNthPrimeRange( arg1, arg2 ):
    n = int( validateRealInt( arg1 ) )
    count = int( validateRealInt( arg2 ) )

    if count < 1:
        return [ ]

    # for primes below 7, we have to do it manually
    if n == 1:
        if count == 1:
            return [ 2 ]
        elif count == 2:
            return[ 2, 3 ]
        elif count == 3:
            return [ 2, 3, 5 ]
        else:
            result = [ 2, 3, 5, 7 ]
            count -= 3
            p = 7
    elif n == 2:
        if count == 1:
            return [ 3 ]
        elif count == 2:
            return [ 3, 5 ]
        else:
            result = [ 3, 5, 7 ]
            count -= 2
            p = 7
    elif n == 3:
        if count == 1:
            return [ 5 ]
        else:
            result = [ 5, 7 ]
            count -= 1
            p = 7
    else:
        p = getNthPrime( n )
        result = [ p ]

    found = 1

    while found < count:
        p = getNextPrimeCandidateForAny( p )

        if isPrime( p ):
            result.append( p )
            found += 1

    return result


#******************************************************************************
#
#  getNthSuperPrime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthSuperPrime( arg ):
    return getNthPrime( getNthPrime( arg ) )


#******************************************************************************
#
#  getNthPolyPrime
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getNthPolyPrime( n, poly ):
    result = getNthPrime( n )

    for _ in arange( 1, poly ):
        result = getNthPrime( result )

    return result


#******************************************************************************
#
#  getPrimes
#
#******************************************************************************

def getPrimes( value, count ):
    for i in getNthPrimeRange( value, count ):
        yield i

@twoArgFunctionEvaluator( )
def getPrimesGenerator( n, k ):
    return RPNGenerator( getPrimes( n, k ) )


#******************************************************************************
#
#  getPrimeRange
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getPrimeRange( start, end ):
    result = list( )

    for i in getNthPrimeRange( start, fadd( fsub( end, start ), 1 ) ):
        result.append( i )

    return result


#******************************************************************************
#
#  getNthPrimorial
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthPrimorial( n ):
    if validateRealInt( n ) == 0:
        return 1

    return fprod( getPrimes( 1, n ) )


#******************************************************************************
#
#  countCache
#
#******************************************************************************

def countCache( name ):
    openPrimeCache( name )
    return g.cursors[ name ].execute( '''SELECT count( id ) FROM cache''' ).fetchone( )[ 0 ]


#******************************************************************************
#
#  getMaxPrime
#
#******************************************************************************

def getMaxPrime( name ):
    openPrimeCache( name )
    return g.cursors[ name ].execute( '''SELECT max( id ), value FROM cache''' ).fetchone( )


#******************************************************************************
#
#  millerRabinPass
#
#  https://gist.github.com/sharnett/5479106
#
#******************************************************************************

def millerRabinPass( a, s, d, n ):
    aToPower = pow( a, d, n )

    if aToPower == 1:
        return True

    for _ in range( s - 1 ):
        if aToPower == n - 1:
            return True

        aToPower = ( aToPower * aToPower ) % n

    return aToPower == n - 1


#******************************************************************************
#
#  isPrimeMillerRabin
#
#  https://gist.github.com/sharnett/5479106
#  http://mathworld.wolfram.com/StrongPseudoprime.html
#
#  http://pi-e.de/Miller-Rabin-Pseudoprimzahlen.htm - check this out!
#
#******************************************************************************

def isPrimeMillerRabin( n ):
    if n < 2:
        return False

    if n < 2047:
        testPrimes = [ 2 ]
    elif n < 1373653:
        testPrimes = [ 2, 3 ]
    elif n < 9080191:
        testPrimes = [ 31, 73 ]
    elif n < 4759123141:
        testPrimes = [ 2, 7, 61 ]
    elif n < 2152302898747:
        testPrimes = [ 2, 3, 5, 7, 11 ]
    elif n < 3474749660383:
        testPrimes = [ 2, 3, 5, 7, 11, 13 ]
    elif n < 341550071728321:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17 ]
    elif n < 18446744073709551616:
        testPrimes = [ 2, 325, 9375, 28178, 450775, 9780504, 1795265022 ]
    elif n < 318665857834031151167461:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37 ]
    elif n < 3317044064679887385961981:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41 ]
    elif n < 6003094289670105800312596501 and g.zhangConjecturesAllowed:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43 ]
    elif n < 59276361075595573263446330101 and g.zhangConjecturesAllowed:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47 ]
    elif n < 564132928021909221014087501701 and g.zhangConjecturesAllowed:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59 ]
    elif n < 1543267864443420616877677640751301 and g.zhangConjecturesAllowed:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67 ]
    else:
        if g.zhangConjecturesAllowed:
            raise ValueError( str( n ) + ' is too large to use the deterministic Miller-Rabin test for 19 prime bases' )

        raise ValueError( str( n ) + ' is too large to use the deterministic Miller-Rabin test for 13 prime bases' )

    if n in testPrimes:
        return True

    d = n - 1
    s = 0

    while d % 2 == 0:
        d >>= 1
        s += 1

    for a in testPrimes:
        if not millerRabinPass( a, s, d, n ):
            return False

    return True


#******************************************************************************
#
#  isStrongPseudoprime
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def isStrongPseudoprime( n, k ):
    if n < k or fmod( n, 2 ) == 0 or isPrime( n ):
        return 0

    d = int( n - 1 )
    s = 0

    while d % 2 == 0:
        d >>= 1
        s += 1

    return 1 if millerRabinPass( int( k ), s, d, int( n ) ) else 0


#******************************************************************************
#
#  checkForPrimeData
#
#******************************************************************************

def checkForPrimeData( ):
    primeFile = Path( getUserDataPath( ) + os.sep + 'large_primes.cache' )

    if primeFile.is_file( ):
        g.primeDataAvailable = True
    else:
        g.primeDataAvailable = False




#!/usr/bin/env python

#******************************************************************************
#
#  rpnKeyboard.py
#
#  rpnChilada keyboard utility functions
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import signal


#******************************************************************************
#
#  class DelayedKeyboardInterrupt
#
#  http://stackoverflow.com/questions/842557/how-to-prevent-a-block-of-code-from-\
#      being-interrupted-by-keyboardinterrupt-in-py
#
#******************************************************************************

class DelayedKeyboardInterrupt( ):
    '''This class is used to mask keyboard interrupts.'''
    def __init__( self ):
        self.signalReceived = False
        self.oldHandler = None

    def __enter__( self ):
        self.signalReceived = False
        self.oldHandler = signal.getsignal( signal.SIGINT )
        signal.signal( signal.SIGINT, self.handler )

    def handler( self, signalType, frame ):
        self.signalReceived = ( signalType, frame )

    def __exit__( self, signalType, value, traceback ):
        signal.signal( signal.SIGINT, self.oldHandler )

        if self.signalReceived:
            self.oldHandler( *self.signalReceived )




#!/usr/bin/env python

#******************************************************************************
#
#  rpnGlobals.py
#
#  rpnChilada global declarations
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

# pylint: disable=invalid-name


# general globals
dataDir = 'rpndata'

# base identifiers
phiBase = -1
fibBase = -2
facBase = -3
doublefacBase = -4
squareBase = -5
lucasBase = -6
triangularBase = -7
primorialBase = -8
eBase = -9
piBase = -10
sqrt2Base = -11

maxSpecialBase = -11

# default options
defaultBitwiseGroupSize = 16
defaultCFTerms = 10
defaultDecimalDelimiter = ' '
defaultDecimalGrouping = 5
defaultInputRadix = 10
defaultIntegerDelimiter = ','
defaultIntegerGrouping = 3
defaultLineLength = 80
defaultListFormatLevel = 1
defaultMinValueForYAFU = 1000000000000000000
defaultMaximumFixed = 5
defaultOutputAccuracy = 12
defaultOutputRadix = 10
defaultPrecision = 20

defaultNumerals = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'

# state variables
astroDataLoaded = False         # Whether or not we've tried to load the astronomy data
astroDataAvailable = False      # If the astronomy data is available (i.e., It's there, _and_ we loaded it.)
checkForSingleResults = False   # This is set true for making help and unit tests, because those should always
                                # return single results
creatingFunction = False        # Whether we're in the process of creating a user-defined function
duplicateOperations = 0
echoArguments = [ ]
helpLoaded = False
interactive = False
lastOperand = 0
nestedListLevel = 0
operandsToRemove = 0
operatorList = False            # whether we are in the process of creating an operator list
                                # (between '(' and ')' operators)
operatorsInList = 0
startTime = 0
testFilter = ''
useMembers = 0

# options
bitwiseGroupSize = 0
comma = False
debugMode = False
decimalDelimiter = defaultDecimalDelimiter
decimalGrouping = 0
echo_command = False
identify = False
ignoreCache = False
inputRadix = defaultInputRadix
integerDelimiter = defaultIntegerDelimiter
integerGrouping = defaultIntegerGrouping
leadingZero = False
lineLength = defaultLineLength
listFormatLevel = 0
maximumFixed = defaultMaximumFixed
maxToFactorByTrialDivision = 100000000  # 10,000 squared because we check all primes below 10,000
minValueForYAFU = defaultMinValueForYAFU
numerals = ''
outputAccuracy = -1
outputRadix = defaultOutputRadix
tempCommaMode = False
tempHexMode = False
tempIdentifyMode = False
tempLeadingZeroMode = False
tempOctalMode = False
tempTimerMode = False
timeIndividualTests = False
timeLimit = 0
timer = False
useYAFU = True
verbose = False
zhangConjecturesAllowed = False

# internal constants
unitConversionPrecision = 50

# unit data
basicUnitTypes = { }
constantOperators = { }
constantOperatorNames = [ ]
aliases = { }
unitConversionMatrix = { }
unitOperators = { }
unitOperatorNames = [ ]

# help data
helpTopics = { }
helpVersion = '0.0.0'
operatorCategories = [ ]
operatorHelp = { }
unitTypeDict = { }

# interactive mode
results = [ ]
variables = { }
promptCount = 0

# factor cache
minValueToCache = 100000
factorCache = None

# location cache
locationCache = None

# user-defined constants
constants = { }

# all keywords
keywords = [ ]

# operator caches
functionCaches = { }

# prime number data
databases = { }
cursors = { }
primeDataAvailable = False

# chemistry tables
elements = None
atomic_numbers = None

# config
userVariables = { }
userVariablesAreDirty = False
userFunctions = { }
userFunctionsAreDirty = False
userConfiguration = { }
userConfigurationIsDirty = False

# astronomy data
planets = None
ephemeris = None
timescale = None

# constants
c = None
e = None
e0 = None
G = None
h = None
h_bar = None
k = None




#!/usr/bin/env python

#******************************************************************************
#
#  rpnDateTime.py
#
#  rpnChilada date and time operations
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import calendar
import datetime
import pytz

import arrow

from dateutil import tz

from mpmath import floor, fmod, fmul, fneg, fsub, nan

from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMeasurement import RPNMeasurement, convertUnits
from rpn.rpnUtils import oneArgFunctionEvaluator, validateRealInt

import rpn.rpnGlobals as g


#******************************************************************************
#
#  month and day constants
#
#******************************************************************************

MONDAY = 1
TUESDAY = 2
WEDNESDAY = 3
THURSDAY = 4
FRIDAY = 5
SATURDAY = 6
SUNDAY = 7

JANUARY = 1
FEBRUARY = 2
MARCH = 3
APRIL = 4
MAY = 5
JUNE = 6
JULY = 7
AUGUST = 8
SEPTEMBER = 9
OCTOBER = 10
NOVEMBER = 11
DECEMBER = 12


#******************************************************************************
#
#  getLocalTimeZone
#
#******************************************************************************

def getLocalTimeZone( ):
    if 'time_zone' in g.userVariables:
        return pytz.timezone( g.userVariables[ 'time_zone' ] )
    elif tz.tzlocal( ) is None:
        return pytz.timezone( 'US/Eastern' )
    else:
        return tz.tzlocal( )


#******************************************************************************
#
#  class RPNDateTime
#
#******************************************************************************

class RPNDateTime( arrow.Arrow ):
    '''This class wraps the Arrow class, with lots of convenience functions and
    implements support for date math.'''
    def __init__( self, year, month, day, hour = 0, minute = 0, second = 0,
                  microsecond = 0, tzinfo = getLocalTimeZone( ), dateOnly = False ):
        self.dateOnly = dateOnly
        super( RPNDateTime, self ).__init__( int( year ), int( month ), int( day ),
                                             int( hour ), int( minute ), int( second ),
                                             int( microsecond ), tzinfo )

    def setDateOnly( self, dateOnly = True ):
        self.dateOnly = dateOnly

    def getDateOnly( self ):
        return self.dateOnly

    @staticmethod
    def get( *args, **kwargs ):
        result = arrow.api.get( *args, **kwargs )

        return RPNDateTime( result.year, result.month, result.day, result.hour,
                            result.minute, result.second, result.microsecond, result.tzinfo )

    def getYMD( self ):
        return ( self.year, self.month, self.day )

    @staticmethod
    def getUTCOffset( timeZone = getLocalTimeZone( ) ):
        dateTime = datetime.datetime.now( timeZone )
        return RPNMeasurement( dateTime.utcoffset( ).total_seconds( ), 'seconds' )

    def getLocalTime( self, timeZone = getLocalTimeZone( ) ):
        result = self
        result = result.add( self.getUTCOffset( timeZone ) )
        #return result.subtract( RPNMeasurement( result.astimezone( tz ).dst( ).seconds, 'seconds' ) )
        return result

    @staticmethod
    def parseDateTime( n ):
        result = arrow.api.get( n )

        return RPNDateTime( result.year, result.month, result.day, result.hour,
                            result.minute, result.second, result.microsecond, result.tzinfo )

    @staticmethod
    def convertFromArrow( arrowDT ):
        return RPNDateTime( arrowDT.year, arrowDT.month, arrowDT.day, arrowDT.hour,
                            arrowDT.minute, arrowDT.second, arrowDT.microsecond, arrowDT.tzinfo )

    @staticmethod
    def convertFromEphemDate( ephemDate ):
        dateValues = list( ephemDate.tuple( ) )

        dateValues.append( int( fmul( fsub( dateValues[ 5 ], floor( dateValues[ 5 ] ) ), 1000000 ) ) )
        dateValues[ 5 ] = int( floor( dateValues[ 5 ] ) )

        return RPNDateTime( *dateValues )

    @staticmethod
    def getNow( ):
        return RPNDateTime.convertFromArrow( arrow.now( ) )

    def compare( self, value ):
        if self.year > value.year:
            return 1
        elif self.year < value.year:
            return -1
        elif self.month > value.month:
            return 1
        elif self.month > value.month:
            return -1
        elif self.day > value.day:
            return 1
        elif self.day > value.day:
            return -1
        elif self.hour > value.hour:
            return 1
        elif self.hour > value.hour:
            return -1
        elif self.minute > value.minute:
            return 1
        elif self.minute > value.minute:
            return -1
        elif self.second > value.second:
            return 1
        elif self.second > value.second:
            return -1
        elif self.microsecond > value.microsecond:
            return 1
        elif self.microsecond > value.microsecond:
            return -1
        else:
            return 0

    def incrementMonths( self, months ):
        newDay = self.day
        newMonth = self.month + int( months )
        newYear = self.year

        if not 1 < newMonth < 12:
            newYear += ( newMonth - 1 ) // 12
            newMonth = ( ( newMonth - 1 ) % 12 ) + 1

        maxDay = calendar.monthrange( newYear, newMonth )[ 1 ]

        if newDay > maxDay:
            newDay = maxDay

        return RPNDateTime( newYear, newMonth, newDay, self.hour, self.minute, self.second )

    def add( self, time ):
        if not isinstance( time, RPNMeasurement ):
            ValueError( 'RPNMeasurement expected' )

        #print( 'time.getUnitName( )', time.getUnitName( ) )
        #print( 'g.unitOperators[ time.getUnitName( ) ].categories', g.unitOperators[ time.getUnitName( ) ].categories )

        if 'years' in g.unitOperators[ time.getUnitName( ) ].categories:
            years = time.convertValue( 'year' )
            return self.replace( year = self.year + years )
        elif 'months' in g.unitOperators[ time.getUnitName( ) ].categories:
            months = time.convertValue( 'month' )
            return self.incrementMonths( months )
        else:
            days = int( floor( time.convertValue( 'day' ) ) )
            seconds = int( fmod( floor( time.convertValue( 'second' ) ), 86400 ) )
            microseconds = int( fmod( floor( time.convertValue( 'microsecond' ) ), 1000000 ) )

            try:
                return self + datetime.timedelta( days = days, seconds = seconds, microseconds = microseconds )
            except OverflowError:
                print( 'rpn:  value is out of range to be converted into a time' )
                return nan

    def format( self ):
        return '{0:4d}-{1:02d}-{2:02d} {3:02d}:{4:02d}:{5:02d}'.format( self.year, self.month, self.day,
                                                                        self.hour, self.minute, self.second )

    def formatDate( self ):
        return '{0:4d}-{1:02d}-{2:02d}'.format( self.year, self.month, self.day )

    def formatTime( self ):
        return '{0:02d}:{1:02d}:{2:02d}'.format( self.hour, self.minute, self.second )

    def subtract( self, time ):
        if isinstance( time, RPNMeasurement ):
            kneg = RPNMeasurement( fneg( time.value ), time.units )
            return self.add( kneg )
        elif isinstance( time, RPNDateTime ):
            if self > time:
                delta = self - time
                factor = 1
            else:
                delta = time - self
                factor = -1

            if delta.days != 0:
                result = RPNMeasurement( delta.days * factor, 'day' )
                result = result.add( RPNMeasurement( delta.seconds * factor, 'second' ) )
                result = result.add( RPNMeasurement( delta.microseconds * factor, 'microsecond' ) )
            elif delta.seconds != 0:
                result = RPNMeasurement( delta.seconds * factor, 'second' )
                result = result.add( RPNMeasurement( delta.microseconds * factor, 'microsecond' ) )
            else:
                result = RPNMeasurement( delta.microseconds * factor, 'microsecond' )

            return result
        else:
            raise ValueError( 'incompatible type for subtracting from an absolute time' )

    def __gt__( self, value ):
        return self.compare( value ) > 0

    def __lt__( self, value ):
        return self.compare( value ) < 0

    def __eq__( self, value ):
        return self.compare( value ) == 0

    def __ge__( self, value ):
        return self.compare( value ) >= 0

    def __le__( self, value ):
        return self.compare( value ) <= 0


#******************************************************************************
#
#  convertToUnixTime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def convertToUnixTime( n ):
    try:
        result = RPNDateTime.parseDateTime( n ).timestamp
    except OverflowError:
        raise ValueError( 'out of range error' )
    except TypeError:
        raise ValueError( 'expected time value' )
    except OSError:
        raise ValueError( 'out of range error' )

    return result


#******************************************************************************
#
#  convertFromUnixTime
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def convertFromUnixTime( n ):
    try:
        result = RPNDateTime.parseDateTime( validateRealInt( n ) )
    except OverflowError:
        raise ValueError( 'out of range error' )
    except TypeError:
        raise ValueError( 'expected time value' )
    except OSError:
        raise ValueError( 'out of range error' )

    return result


#******************************************************************************
#
#  convertToHMS
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def convertToHMS( n ):
    return convertUnits( n, [ 'hour', 'minute', 'second' ] )


#******************************************************************************
#
#  convertToDHMS
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def convertToDHMS( n ):
    return convertUnits( n, [ 'day', 'hour', 'minute', 'second' ] )


#******************************************************************************
#
#  convertToYDHMS
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def convertToYDHMS( n ):
    return convertUnits( n, [ 'year', 'day', 'hour', 'minute', 'second' ] )


#******************************************************************************
#
#  makeJulianTime
#
#******************************************************************************

def makeJulianTime( n ):
    if isinstance( n, RPNGenerator ):
        return makeJulianTime( list( n ) )
    elif len( n ) == 1:
        return RPNDateTime( n[ 0 ], 1, 1 )

    result = RPNDateTime( n[ 0 ], 1, 1 ).add( RPNMeasurement( n[ 1 ] - 1, 'day' ) )

    if len( n ) >= 3:
        result = result.replace( hour = n[ 2 ] )

    if len( n ) >= 4:
        result = result.replace( minute = n[ 3 ] )

    if len( n ) >= 5:
        result = result.replace( second = n[ 4 ] )

    if len( n ) >= 6:
        result = result.replace( microsecond = n[ 5 ] )

    return result


#******************************************************************************
#
#  makeISOTime
#
#******************************************************************************

def makeISOTime( n ):
    if isinstance( n, RPNGenerator ):
        return makeISOTime( list( n ) )
    elif len( n ) == 1:
        year = n[ 0 ]
        week = 1
        day = 1
    elif len( n ) == 2:
        year = n[ 0 ]
        week = n[ 1 ]
        day = 1
    else:
        year = n[ 0 ]
        week = n[ 1 ]
        day = n[ 2 ]

    result = datetime.datetime.strptime( '%04d-%02d-%1d' % ( int( year ), int( week ), int( day ) ), '%Y-%W-%w' )

    if RPNDateTime( year, 1, 4 ).isoweekday( ) > 4:
        result -= datetime.timedelta( days = 7 )

    return result


#******************************************************************************
#
#  makeDateTime
#
#******************************************************************************

def makeDateTime( n ):
    if isinstance( n, ( RPNGenerator, int ) ):
        return makeDateTime( list( n ) )
    elif isinstance( n, str ):
        return RPNDateTime.get( n )
    elif isinstance( n[ 0 ], list ):
        return [ makeDateTime( i ) for i in n ]

    if len( n ) == 1:
        n.append( 1 )

    if len( n ) == 2:
        n.append( 1 )
    elif len( n ) > 7:
        n = n[ : 7 ]

    return RPNDateTime( *n )


#******************************************************************************
#
#  getNow
#
#******************************************************************************

def getNow( ):
    return RPNDateTime.now( tzinfo = getLocalTimeZone( ) )


#******************************************************************************
#
#  getToday
#
#******************************************************************************

def getToday( ):
    now = getNow( )
    return RPNDateTime( *now.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  getTomorrow
#
#******************************************************************************

def getTomorrow( ):
    now = getNow( )
    now = now + datetime.timedelta( days = 1 )
    return RPNDateTime( *now.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  getYesterday
#
#******************************************************************************

def getYesterday( ):
    now = getNow( )
    now = now + datetime.timedelta( days = -1 )
    return RPNDateTime( *now.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  calculateEaster
#
#  This algorithm comes from Gauss.
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateEaster( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    a = year % 19
    b = year // 100
    c = year % 100
    d = ( 19 * a + b - b // 4 - ( ( b - ( b + 8 ) // 25 + 1 ) // 3 ) + 15 ) % 30
    e = ( 32 + 2 * ( b % 4 ) + 2 * ( c // 4 ) - d - ( c % 4 ) ) % 7
    f = d + e - 7 * ( ( a + 11 * d + 22 * e ) // 451 ) + 114
    month = f // 31
    day = f % 31 + 1

    return RPNDateTime( year, month, day, dateOnly = True )


#******************************************************************************
#
#  calculateAshWednesday
#
#  46 days before Easter (40 days, not counting Sundays)
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateAshWednesday( year ):
    ashWednesday = calculateEaster( validateRealInt( year ) ).add( RPNMeasurement( -46, 'day' ) )
    return RPNDateTime( *ashWednesday.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  calculateGoodFriday
#
#  2 days before Easter
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateGoodFriday( year ):
    goodFriday = calculateEaster( validateRealInt( year ) ).add( RPNMeasurement( -2, 'day' ) )
    return RPNDateTime( *goodFriday.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  getLastDayOfMonth
#
#******************************************************************************

def getLastDayOfMonth( year, month ):
    return calendar.monthrange( validateRealInt( year ), validateRealInt( month ) )[ 1 ]


#******************************************************************************
#
#  calculateNthWeekdayOfYear
#
#  Monday = 1, etc., as per arrow, nth == -1 for last
#
#******************************************************************************

def calculateNthWeekdayOfYear( year, nth, weekday ):
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    if validateRealInt( nth ) > 0:
        firstDay = RPNDateTime( year, 1, 1 )

        firstWeekDay = validateRealInt( weekday ) - firstDay.isoweekday( ) + 1

        if firstWeekDay < 1:
            firstWeekDay += 7

        result = RPNDateTime( year, 1, firstWeekDay ).add( RPNMeasurement( nth - 1, 'week' ) )
        result.setDateOnly( )

        return result
    elif nth < 0:
        lastDay = RPNDateTime( year, 12, 31 )

        lastWeekDay = validateRealInt( weekday ) - lastDay.isoweekday( )

        if lastWeekDay > 0:
            lastWeekDay -= 7

        lastWeekDay += 31

        result = RPNDateTime( year, 12, lastWeekDay, dateOnly = True ).add( RPNMeasurement( ( nth + 1 ), 'week' ) )
        result.setDateOnly( )

        return result

    raise ValueError( '0th weekday makes no sense in this context' )


#******************************************************************************
#
#  calculateNthWeekdayOfMonth
#
#  Monday = 1, etc.
# ( -1 == last. -2 == next to last, etc.)
#
#******************************************************************************

def calculateNthWeekdayOfMonth( year, month, nth, weekday ):
    if validateRealInt( weekday ) > SUNDAY or weekday < MONDAY:
        raise ValueError( 'day of week must be 1 - 7 (Monday to Sunday)' )

    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    firstDayOfWeek = arrow.Arrow( validateRealInt( year ), validateRealInt( month ), 1 ).isoweekday( )

    if nth < 0:
        day = ( ( weekday + 1 ) - firstDayOfWeek ) % 7

        while day <= getLastDayOfMonth( year, month ):
            day += 7

        day += nth * 7
    else:
        day = ( weekday - firstDayOfWeek + 1 ) + nth * 7

        if weekday >= firstDayOfWeek:
            day -= 7

    return RPNDateTime( year, month, day, dateOnly = True )


#******************************************************************************
#
#  calculateThanksgiving
#
#  the fourth Thursday in November
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateThanksgiving( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    return calculateNthWeekdayOfMonth( year, NOVEMBER, 4, THURSDAY )


#******************************************************************************
#
#  calculateLaborDay
#
#  the first Monday in September
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateLaborDay( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    return calculateNthWeekdayOfMonth( year, SEPTEMBER, 1, MONDAY )


#******************************************************************************
#
#  calculateElectionDay
#
#  the first Tuesday after the first Monday (so it's never on the 1st day)
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateElectionDay( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    result = calculateNthWeekdayOfMonth( year, NOVEMBER, 1, MONDAY )
    result.replace( day = result.day + 1 )

    return RPNDateTime( *result.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  calculateMemorialDay
#
#  the last Monday in May (4th or 5th Monday)
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateMemorialDay( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    return calculateNthWeekdayOfMonth( year, MAY, -1, MONDAY )


#******************************************************************************
#
#  calculatePresidentsDay
#
#  the third Monday in February
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculatePresidentsDay( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    return calculateNthWeekdayOfMonth( year, FEBRUARY, 3, MONDAY )


#******************************************************************************
#
#  calculateMartinLutherKingDay
#
#  the third Monday in January
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateMartinLutherKingDay( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    return calculateNthWeekdayOfMonth( year, JANUARY, 3, MONDAY )


#******************************************************************************
#
#  calculateColumbusDay
#
#  the second Monday in October
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateColumbusDay( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    return calculateNthWeekdayOfMonth( year, OCTOBER, 2, MONDAY )


#******************************************************************************
#
#  calculateMothersDay
#
#  the second Sunday in May (in the U.S. and most other countries)
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateMothersDay( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    return calculateNthWeekdayOfMonth( year, MAY, 2, SUNDAY )


#******************************************************************************
#
#  calculateFathersDay
#
#  the third Sunday in June (in the U.S. and most other countries)
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateFathersDay( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    return calculateNthWeekdayOfMonth( year, JUNE, 3, SUNDAY )


#******************************************************************************
#
#  getNewYearsDay
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNewYearsDay( year ):
    return RPNDateTime( year, JANUARY, 1, dateOnly = True )


#******************************************************************************
#
#  getVeteransDay
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getVeteransDay( year ):
    return RPNDateTime( year, NOVEMBER, 11, dateOnly = True )


#******************************************************************************
#
#  getIndependenceDay
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getIndependenceDay( year ):
    return RPNDateTime( year, JULY, 4, dateOnly = True )


#******************************************************************************
#
#  getChristmasDay
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getChristmasDay( year ):
    return RPNDateTime( year, DECEMBER, 25, dateOnly = True )


#******************************************************************************
#
#  calculateAdvent
#
#  4 Sundays before Christmas
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateAdvent( year ):
    firstAdvent = getChristmasDay( validateRealInt( year ) ).add( RPNMeasurement( -3, 'week' ) )
    firstAdvent = firstAdvent.subtract( RPNMeasurement( getWeekday( firstAdvent ), 'day'  ) )

    return RPNDateTime( *firstAdvent.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  getEpiphanyDay
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getEpiphanyDay( year ):
    return RPNDateTime( year, 1, 6, dateOnly = True )


#******************************************************************************
#
#  calculatePentecostSunday
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculatePentecostSunday( year ):
    return RPNDateTime( *( calculateEaster( year ).add( RPNMeasurement( 7, 'weeks' ) ).getYMD( ) ),
                        dateOnly = True )


#******************************************************************************
#
#  calculateAscensionThursday
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateAscensionThursday( year ):
    '''
I don't know why it's 39 days after Easter instead of 40, but that's how
the math works out.  It's the 40th day of the Easter season.

Or as John Wright says, "Catholics can't count."
    '''
    return RPNDateTime( *calculateEaster( year ).add( RPNMeasurement( 39, 'days' ) ).getYMD( ),
                        dateOnly = True )


#******************************************************************************
#
#  calculateDSTStart
#
#  the second Sunday in March
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateDSTStart( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    if year >= 2007:
        return calculateNthWeekdayOfMonth( year, MARCH, 2, SUNDAY )
    elif year == 1974:
        return RPNDateTime( 1974, JANUARY, 7, dateOnly = True )
    elif year >= 1967:
        return calculateNthWeekdayOfMonth( year, APRIL, 1, SUNDAY )
    else:
        raise ValueError( 'DST was not standardized before 1967' )


#******************************************************************************
#
#  calculateDSTEnd
#
#  the first Sunday in November
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def calculateDSTEnd( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = validateRealInt( year )

    if year >= 2007:
        return calculateNthWeekdayOfMonth( year, NOVEMBER, 1, SUNDAY )
    elif year == 1974:
        return RPNDateTime( 1974, DECEMBER, 31, dateOnly = True )  # technically DST never ended in 1974
    elif year >= 1967:
        return calculateNthWeekdayOfMonth( year, OCTOBER, -1, SUNDAY )
    else:
        raise ValueError( 'DST was not standardized before 1967' )


#******************************************************************************
#
#  getISODay
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getISODay( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'date/time type required for this operator' )

    return list( n.isocalendar( ) )


#******************************************************************************
#
#  getWeekday
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getWeekday( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'date/time type required for this operator' )

    return n.weekday( ) + 1


#******************************************************************************
#
#  getWeekdayName
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getWeekdayName( n ):
    return calendar.day_name[ getWeekday( n ) - 1 ]


#******************************************************************************
#
#  getYear
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getYear( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'date/time type required for this operator' )

    return n.year


#******************************************************************************
#
#  getMonth
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getMonth( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'date/time type required for this operator' )

    return n.month


#******************************************************************************
#
#  getDay
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getDay( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'date/time type required for this operator' )

    return n.day


#******************************************************************************
#
#  getHour
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getHour( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'date/time type required for this operator' )

    return n.hour


#******************************************************************************
#
#  getMinute
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getMinute( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'date/time type required for this operator' )

    return n.minute


#******************************************************************************
#
#  getSecond
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getSecond( n ):
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'date/time type required for this operator' )

    return n.second + n.microsecond / 1000000


#******************************************************************************
#
#  isDST
#
#******************************************************************************

def isDST( dateTime, timeZone ):
    return dateTime.astimezone( timeZone ).dst( ) != datetime.timedelta( 0 )




#!/usr/bin/env python

#******************************************************************************
#
#  rpnName.py
#
#  rpnChilada functions for converting integers to English names
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

from mpmath import fabs, fdiv, floor, fmod, power

from rpn.rpnMeasurement import RPNMeasurement
from rpn.rpnUtils import oneArgFunctionEvaluator, validateRealInt


#******************************************************************************
#
#  getModifiedOnesName
#
#******************************************************************************

def getModifiedOnesName( name, code ):
    if ( 'n' in code ) and ( name in [ 'septe', 'nove' ] ):
        return name + 'n'

    if ( 'm' in code ) and ( name in [ 'septe', 'nove' ] ):
        return name + 'm'

    if ( 's' in code ) and ( name in [ 'tre', 'se' ] ):
        return name + 's'

    if 'x' in code:
        if name == 'tre':
            return name + 's'

        if name == 'se':
            return name + 'x'

        return name

    return name


#******************************************************************************
#
#  getSmallNumberName
#
#  Returns an english number name for anything from 0 to 999.
#
#******************************************************************************

def getSmallNumberName( n, ordinal = False ):
    unitNumberNames = \
        [ '', 'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight',
          'nine', 'ten', 'eleven', 'twelve', 'thirteen', 'fourteen', 'fifteen',
          'sixteen', 'seventeen', 'eighteen', 'nineteen' ]

    unitOrdinalNumberNames = \
        [ '', 'first', 'second', 'third', 'fourth', 'fifth', 'sixth', 'seventh',
          'eighth', 'ninth', 'tenth', 'eleventh', 'twelfth', 'thirteenth',
          'fourteenth', 'fifteenth', 'sixteenth', 'seventeenth', 'eighteenth',
          'nineteenth' ]

    tensNumberNames = \
        [ '', '', 'twenty', 'thirty', 'forty', 'fifty', 'sixty', 'seventy',
          'eighty', 'ninety' ]

    tensOrdinalNumberNames = \
        [ '', '', 'twentieth', 'thirtieth', 'fortieth',
          'fiftieth', 'sixtieth', 'seventieth', 'eightieth', 'ninetieth' ]

    hundreds = n // 100
    tens = ( n // 10 ) % 10
    ones = n % 10

    name = ''

    if hundreds > 0:
        name = unitNumberNames[ hundreds ] + ' hundred'

        if ordinal and tens == 0 and ones == 0:
            name += 'th'

    if tens > 1:
        if name != '':
            name += ' '

        if ordinal and ones == 0:
            name += tensOrdinalNumberNames[ tens ]
        else:
            name += tensNumberNames[ tens ]

    if ones > 0:
        if tens > 1:
            name += '-'

            if ordinal:
                name += unitOrdinalNumberNames[ ones ]
            else:
                name += unitNumberNames[ ones ]
        elif tens == 1:
            if name != '':
                name += ' '

            if ordinal:
                name += unitOrdinalNumberNames[ ones + 10 ]
            else:
                name += unitNumberNames[ ones + 10 ]
        else:
            if name != '':
                name += ' '

            if ordinal:
                name += unitOrdinalNumberNames[ ones ]
            else:
                name += unitNumberNames[ ones ]
    elif tens == 1:
        if name != '':
            name += ' '

        if ordinal:
            name += unitOrdinalNumberNames[ 10 ]
        else:
            name += unitNumberNames[ 10 ]

    return name


#******************************************************************************
#
#  getNumberGroupName
#
#  returns the name of the "group", i.e., the three-digit group of an integer
#  separated by commas in the usual notation.  Group 0 has no name because it
#  represents the "ones".  Group 1 is "thousand", group 2 is "million", etc.
#
#******************************************************************************

def getNumberGroupName( n ):
    groupNames = [ '', 'thousand', 'million', 'billion', 'trillion',
                   'quadrillion', 'quintillion', 'sextillion', 'septillion',
                   'octillion', 'nonillion', 'decillion' ]

    onesNames = [ '', 'un', 'duo', 'tre', 'quattuor', 'quinqua', 'se', 'septe', 'octo', 'nove' ]

    tensNames = [ ( '', '' ), ( 'deci', 'n' ), ( 'viginti', 'ms' ), ( 'triginta', 'ns' ),
                  ( 'quadraginta', 'ns' ), ( 'quinquaginta', 'ns' ), ( 'sexaginta', 'n' ),
                  ( 'septuaginta', 'n' ), ( 'octoginta', 'mx' ), ( 'nonaginta', '' ) ]

    hundredsNames = [ ( '', '' ), ( 'centi', 'nx' ), ( 'ducenti', 'n' ), ( 'trecenti', 'ns' ),
                      ( 'quadringenti', 'ns' ), ( 'quingenti', 'ns' ), ( 'sescenti', 'n' ),
                      ( 'septingenti', 'n' ), ( 'octingenti', 'mx' ), ( 'nongenti', '' ) ]

    if n < len( groupNames ):
        return groupNames[ n ]
    else:
        n -= 1

        hundreds = n // 100
        tens = ( n // 10 ) % 10
        ones = n % 10

        name = ''
        hasTens = False

        if ones > 0:
            name += onesNames[ ones ]

        if tens > 0:
            hasTens = True

            if name != '':
                name = getModifiedOnesName( name, tensNames[ tens ][ 1 ] )

            name += tensNames[ tens ][ 0 ]

        if hundreds > 0:
            if not hasTens:
                name = getModifiedOnesName( name, hundredsNames[ hundreds ][ 1 ] )

            name += hundredsNames[ hundreds ][ 0 ]

        if name[ -1 ] in 'ai':
            name = name[ : -1 ]

        name += 'illion'

        return name


#******************************************************************************
#
#  getShortOrdinalName
#
#******************************************************************************

def getShortOrdinalName( n ):
    if n == 11:
        return '11th'
    elif n == 12:
        return '12th'
    elif n == 13:
        return '13th'

    modulo = int( fmod( n, 10 ) )

    if modulo in [ 0, 4, 5, 6, 7, 8, 9 ]:
        result = str( int( n ) ) + 'th'
    elif modulo == 1:
        result = str( int( n ) ) + 'st'
    elif modulo == 2:
        result = str( int( n ) ) + 'nd'
    else:
        result = str( int( n ) ) + 'rd'

    length = len( result )
    offset = -5

    while offset > -length:
        result = result[ : offset ] + ',' + result[ offset : ]
        offset -= 4
        length += 1

    return result


#******************************************************************************
#
#  getOrdinalName
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getOrdinalName( n ):
    return getNumberName( validateRealInt( n ), ordinal = True )


#******************************************************************************
#
#  getNumberName
#
#******************************************************************************

def getNumberName( n, ordinal = False ):
    units = ''

    if isinstance( n, RPNMeasurement ):
        value = n.value

        if value in [ 1, -1 ]:
            units = n.getUnitName( )
        else:
            units = n.getPluralUnitName( )

        n = validateRealInt( value )

    if n == 0:
        if ordinal:
            name = 'zeroth'
        else:
            name = 'zero'

        if units:
            name += ' '
            name += units

        return name

    current = fabs( n )

    if current >= power( 10, 3003 ):
        raise ValueError( 'value out of range for converting to an English name' )

    group = 0
    name = ''

    firstTime = True

    while current > 0:
        section = getSmallNumberName( int( fmod( current, 1000 ) ), ordinal if firstTime else False )

        firstTime = False

        if section != '':
            groupName = getNumberGroupName( group )

            if groupName != '':
                section += ' ' + groupName

                if ordinal and name == '':
                    section += 'th'

            if name == '':
                name = section
            else:
                name = section + ' ' + name

        current = floor( fdiv( current, 1000 ) )
        group += 1

    if n < 0:
        name = 'negative ' + name

    if units:
        name += ' '
        name += units

    return name


@oneArgFunctionEvaluator( )
def getName( n ):
    return getNumberName( n )




#!/usr/bin/env python

#******************************************************************************
#
#  rpnFactor.py
#
#  rpnChilada factoring utilities
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import collections
import os
import subprocess

from mpmath import fdiv, floor, fmod, fneg, fprod, log10, mpmathify

from rpn.factorise import factorise
from rpn.rpnDebug import debugPrint
from rpn.rpnPersistence import loadFactorCache
from rpn.rpnPrimes import primes
from rpn.rpnPrimeUtils import isPrime
from rpn.rpnUtils import oneArgFunctionEvaluator

import rpn.rpnGlobals as g


#******************************************************************************
#
#  getFactors
#
#******************************************************************************

def getFactors( target ):
    if target < -1:
        result = [ -1 ]
        result.extend( getFactors( fneg( target ) ) )
        return result
    elif target == -1:
        return [ -1 ]
    elif target == 0:
        return [ 0 ]
    elif target == 1:
        return [ 1 ]

    n = int( floor( target ) )

    if g.factorCache is None:
        loadFactorCache( )

    if not g.ignoreCache:
        if n in g.factorCache:
            if g.verbose and n != 1:
                print( 'cache hit:', n )
                print( )

            debugPrint( '\nfactor cache', n, g.factorCache[ n ] )

            return g.factorCache[ n ]

    try:
        result = factorByTrialDivision( n )   # throws if n is too big

        if n > g.minValueToCache and n not in g.factorCache:
            g.factorCache[ n ] = result

        return result

    except ValueError:
        pass

    if g.useYAFU and n > g.minValueForYAFU:
        result = runYAFU( n )
    else:
        result = factorise( int( n ) )

    if n > g.minValueToCache:
        if g.ignoreCache or ( not g.ignoreCache and n not in g.factorCache ):
            debugPrint( '\ncaching factors of ', n, result )
            g.factorCache[ n ] = result

    return result

@oneArgFunctionEvaluator( )
def getFactorsOperator( n ):
    return getFactors( n )


#******************************************************************************
#
#  getFactorList
#
#******************************************************************************

def getFactorList( n ):
    return list( collections.Counter( getFactors( n ) ).items( ) )


#******************************************************************************
#
#  factorByTrialDivision
#
#  This is sufficiently fast for small numbers... at least for now.
#
#******************************************************************************

def factorByTrialDivision( n ):
    if n > g.maxToFactorByTrialDivision:
        raise ValueError( 'value', n, 'is too big to factor by trial division' )

    result = [ ]

    for i in primes:
        while n > 1 and fmod( n, i ) == 0:
            n = fdiv( n, i )
            result.append( i )

        if isPrime( n ):
            break

    if n > 1:
        result.append( int( n ) )

    return result


#******************************************************************************
#
#  runYAFU
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def runYAFU( n ):
    fullOut = subprocess.run( [ g.userConfiguration[ 'yafu_path' ] + os.sep +
                                g.userConfiguration[ 'yafu_binary' ], '-xover', '120' ],
                              input='factor(' + str( int( n ) ) + ')\n', encoding='ascii', stdout=subprocess.PIPE,
                              cwd=g.userConfiguration[ 'yafu_path' ], check=True ).stdout

    #print( 'out', fullOut )

    out = fullOut[ fullOut.find( '***factors found***' ) : ]

    if len( out ) < 2:
        if log10( n ) > 40:
            raise ValueError( 'yafu seems to have crashed trying to factor ' + str( int( n ) ) +
                              '.\n\nyafu output follows:\n' + fullOut )

        debugPrint( 'yafu seems to have crashed, switching to built-in factoring code' )
        return factorise( int( n ) )

    result = [ ]

    while True:
        prime = ''

        out = out[ out.find( 'P' ) : ]

        if len( out ) < 2:
            break

        out = out[ out.find( '=' ) : ]

        index = 2

        while out[ index ] >= '0' and out[ index ] <= '9':
            prime += out[ index ]
            index += 1

        result.append( mpmathify( prime ) )

    if not result:
        raise ValueError( 'yafu seems to have failed.' )

    answer = fprod( result )

    if answer != n:
        debugPrint( '\nyafu has barfed' )
        for i in result:
            n = fdiv( n, i )

        result.extend( runYAFU( n ) )

    return sorted( result )

